# =============================================================================
# Minet Google Sheets Functions
# =============================================================================
#
# Functions related to google sheets.
#
import browser_cookie3
from ural import is_url
from ural.google import parse_google_drive_url, GoogleDriveFile

from minet.web import request
from minet.cookies import CookieResolver
from minet.constants import COOKIE_BROWSERS
from minet.google.exceptions import (
    GoogleSheetsInvalidTargetError,
    GoogleSheetsInvalidContentTypeError,
    GoogleSheetsMissingCookieError,
    GoogleSheetsNotFoundError,
    GoogleSheetsUnauthorizedError,
    GoogleSheetsMaxAttemptsExceeded,
)


def append_authuser(url, authuser):
    url += "&authuser=%i" % authuser
    return url


def export_google_sheets_as_csv(
    url, cookie=None, authuser=None, max_authuser_attempts=4
):
    if is_url(url):
        parsed = parse_google_drive_url(url)

        if parsed is None or parsed.type != "spreadsheets":
            raise GoogleSheetsInvalidTargetError
    else:
        parsed = GoogleDriveFile("spreadsheets", url)

    base_export_url = parsed.get_export_url()
    export_url = base_export_url

    if authuser is not None:
        if not isinstance(authuser, int) or authuser < 0:
            raise TypeError("authuser should be an int >= 0")

        export_url = append_authuser(export_url, authuser)
        max_authuser_attempts = 1
    else:
        authuser = 0

    if cookie is not None and cookie in COOKIE_BROWSERS:
        jar = getattr(browser_cookie3, cookie)()
        resolver = CookieResolver(jar)
        cookie = resolver(export_url)

        if cookie is None:
            raise GoogleSheetsMissingCookieError

    attempts = max_authuser_attempts

    while True:
        attempts -= 1

        response = request(export_url, cookie=cookie)

        if response.status == 404:
            raise GoogleSheetsNotFoundError

        if response.status == 401:
            raise GoogleSheetsUnauthorizedError

        if response.status == 403:
            authuser += 1

            if attempts != 0:
                export_url = append_authuser(base_export_url, authuser)
                continue

            raise GoogleSheetsMaxAttemptsExceeded

        if "csv" not in response.headers.get("Content-Type", "").lower():
            raise GoogleSheetsInvalidContentTypeError

        break

    return response.text()



# =============================================================================
# Minet Twitter Users CLI Action
# =============================================================================
#
# Logic of the `tw users` action.
#
from twitwi import normalize_user, format_user_as_csv_row
from twitter import TwitterHTTPError
from twitwi.constants import USER_FIELDS, USER_PARAMS
from ebbe import as_chunks

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.exceptions import FatalError
from minet.cli.twitter.utils import (
    is_not_user_id,
    is_probably_not_user_screen_name,
    with_twitter_client,
)


@with_enricher_and_loading_bar(
    headers=USER_FIELDS, title="Retrieving users", unit="users"
)
@with_twitter_client()
def action(cli_args, client, enricher, loading_bar):
    def call_client(users):
        kwargs = {}

        if not cli_args.v2:
            if cli_args.ids:
                kwargs["user_id"] = users
            else:
                kwargs["screen_name"] = users

            return client.call(["users", "lookup"], **kwargs)
        else:
            kwargs["params"] = USER_PARAMS
            route = ["users"]

            if cli_args.ids:
                kwargs["ids"] = users
            else:
                kwargs["usernames"] = users
                route = ["users", "by"]

            return client.call(route, **kwargs)

    for chunk in as_chunks(100, enricher.cells(cli_args.column, with_rows=True)):
        users = ",".join(row[1].lstrip("@") for row in chunk)

        with loading_bar.step(count=len(chunk)):
            for _, user in chunk:
                if cli_args.ids:
                    if is_not_user_id(user):
                        raise FatalError(
                            "The column given as argument doesn't contain user ids, you have probably given user screen names as argument instead. \nTry removing --ids from the command."
                        )

                    key = "id"

                else:
                    if is_probably_not_user_screen_name(user):
                        raise FatalError(
                            "The column given as argument probably doesn't contain user screen names, you have probably given user ids as argument instead. \nTry adding --ids to the command."
                        )
                        # force flag to add

                    key = "screen_name"

            try:
                result = call_client(users)
            except TwitterHTTPError as e:
                if e.e.code == 404:
                    for row, user in chunk:
                        enricher.writerow(row)
                else:
                    raise e

                continue

            users_raw_data = result if not cli_args.v2 else result["data"]

            indexed_result = {}

            for user in users_raw_data:
                user = normalize_user(user, locale=cli_args.timezone, v2=cli_args.v2)
                user_row = format_user_as_csv_row(user)
                indexed_result[user[key]] = user_row

            for row, user in chunk:
                user_row = indexed_result.get(user.lstrip("@"))

                enricher.writerow(row, user_row)



# =============================================================================
# Minet Mediacloud Topic CLI Action
# =============================================================================
#
# Logic of the `mc topic` action.
#
import casanova

from minet.cli.loading_bar import LoadingBar
from minet.mediacloud import MediacloudAPIClient
from minet.mediacloud.types import MediacloudTopicStory


def action(cli_args):
    writer = casanova.writer(cli_args.output, fieldnames=MediacloudTopicStory)

    with LoadingBar(title="Fetching stories", unit="stories") as loading_bar:
        client = MediacloudAPIClient(cli_args.token)

        iterator = client.topic_stories(
            cli_args.topic_id,
            media_id=cli_args.media_id,
            from_media_id=cli_args.from_media_id,
        )

        for story in iterator:
            writer.writerow(story.as_csv_row())
            loading_bar.advance()



# =============================================================================
# Minet Web Utilities
# =============================================================================
#
# Miscellaneous web-related functions used throughout the library.
#
from typing import (
    Optional,
    Callable,
    Tuple,
    Union,
    OrderedDict,
    List,
    Mapping,
    Any,
    Dict,
    Container,
)
from minet.types import AnyTimeout, Redirection, RedirectionStack, HTTPHeaderDict

import re
import cgi
import certifi
import random
import urllib3
import urllib3.exceptions as urllib3_exceptions
import urllib.error
import ural
import json
import mimetypes
import functools
import threading
from http.cookiejar import CookieJar
from bs4 import SoupStrainer
from datetime import datetime
from timeit import default_timer as timer
from io import BytesIO
from threading import Event
from urllib.parse import urljoin, quote
from urllib.request import Request
from urllib3.util.ssl_ import create_urllib3_context
from urllib3.util.request import ACCEPT_ENCODING
from ebbe import rcompose, noop, format_filesize, format_repr
from tenacity import (
    Retrying,
    RetryCallState,
    retry_if_exception,
    stop_after_attempt,
    stop_when_event_set,
    sleep_using_event,
)
from tenacity.wait import wait_base

from minet.scrape.regex import (
    extract_canonical_link,
    extract_javascript_relocation,
    extract_meta_refresh,
)
from minet.scrape.soup import suppress_xml_parsed_as_html_warnings, WonderfulSoup
from minet.encodings import infer_encoding
from minet.loggers import sleepers_logger
from minet.utils import is_binary_mimetype
from minet.cookies import dict_to_cookie_string
from minet.headers import parse_http_refresh, get_encoding_from_content_type_header
from minet.exceptions import (
    RedirectError,
    MaxRedirectsError,
    InfiniteRedirectsError,
    InvalidRedirectError,
    BadlyEncodedLocationHeaderError,
    InvalidURLError,
    InvalidStatusError,
    SelfRedirectError,
    CancelledRequestError,
    FinalTimeoutError,
    PycurlNotInstalledError,
    PycurlError,
    PycurlProtocolError,
    PycurlTimeoutError,
)
from minet.constants import (
    DEFAULT_SPOOFED_TLS_CIPHERS,
    DEFAULT_URLLIB3_TIMEOUT,
    DEFAULT_SPOOFED_UA,
    REDIRECT_STATUSES,
)

PYCURL_SUPPORT = False

try:
    from minet.pycurl import request_with_pycurl

    PYCURL_SUPPORT = True
except ImportError:
    pass

mimetypes.init()

# Patterns
HTML_RE = re.compile(
    rb"^\s*<(?:html|head|body|title|meta|link|span|div|img|ul|ol|[ap!?])", flags=re.I
)

# Constants
CONTENT_PREBUFFER_UP_TO = 1024
STREAMING_CHUNK_SIZE: int = 2**12
LARGE_CONTENT_PREBUFFER_UP_TO = 2**16
EXPECTED_WEB_ERRORS = (
    urllib.error.URLError,
    urllib3_exceptions.HTTPError,
    RedirectError,
    InvalidURLError,
    InvalidStatusError,
    FinalTimeoutError,
    PycurlError,
)

assert CONTENT_PREBUFFER_UP_TO < LARGE_CONTENT_PREBUFFER_UP_TO


def looks_like_html(html_chunk: bytes) -> bool:
    return HTML_RE.match(html_chunk) is not None


def create_pool_manager(
    proxy: Optional[str] = None,
    parallelism: int = 1,
    num_pools: Optional[int] = None,
    insecure: bool = False,
    spoof_tls_ciphers: bool = False,
    **kwargs
) -> urllib3.PoolManager:
    """
    Helper function returning a urllib3 pool manager with sane defaults.
    """

    manager_kwargs: Dict[str, Any] = {"timeout": DEFAULT_URLLIB3_TIMEOUT}

    if not insecure:
        manager_kwargs["cert_reqs"] = "CERT_REQUIRED"
        manager_kwargs["ca_certs"] = certifi.where()
    else:
        manager_kwargs["cert_reqs"] = "CERT_NONE"
        # manager_kwargs['assert_hostname'] = False

        urllib3.disable_warnings()

    manager_kwargs["maxsize"] = parallelism

    if num_pools is not None:
        manager_kwargs["num_pools"] = num_pools

    manager_kwargs.update(kwargs)

    if spoof_tls_ciphers:
        manager_kwargs["ssl_context"] = create_urllib3_context(
            ciphers=DEFAULT_SPOOFED_TLS_CIPHERS
        )

    if proxy is not None:
        proxy = ural.ensure_protocol(proxy)
        return urllib3.ProxyManager(proxy, **manager_kwargs)

    return urllib3.PoolManager(**manager_kwargs)


DEFAULT_POOL_MANAGER = create_pool_manager()


def timeout_to_final_time(timeout: AnyTimeout) -> float:
    seconds: float

    if isinstance(timeout, urllib3.Timeout):
        if timeout.total is not None:
            seconds = timeout.total
        else:
            seconds = timeout.connect_timeout + timeout.read_timeout
    else:
        seconds = timeout

    # Some epsilon so sockets can timeout themselves properly
    seconds += 0.01

    return timer() + seconds


def pool_manager_aware_timeout_to_final_time(
    pool_manager: urllib3.PoolManager, timeout: Optional[AnyTimeout]
) -> Optional[float]:
    if timeout is not None:
        return timeout_to_final_time(timeout)

    else:
        pool_manager_timeout = pool_manager.connection_pool_kw.get("timeout")

        if pool_manager_timeout is not None:
            return timeout_to_final_time(pool_manager_timeout)

    return None


def stream_response_body(
    response: urllib3.HTTPResponse,
    body: BytesIO,
    chunk_size: int = STREAMING_CHUNK_SIZE,
    cancel_event: Optional[Event] = None,
    final_time: Optional[float] = None,
    up_to: Optional[int] = None,
) -> bool:
    if cancel_event is not None and cancel_event.is_set():
        raise CancelledRequestError

    if up_to is not None and body.tell() >= up_to:
        return False

    while True:
        data = response.read(chunk_size)

        if not data:
            break

        body.write(data)

        if up_to is not None and body.tell() >= up_to:
            return False

        if cancel_event is not None and cancel_event.is_set():
            raise CancelledRequestError

        if final_time is not None:
            if timer() >= final_time:
                raise FinalTimeoutError

    # This is the only place we know the body has been fully read
    return True


class BufferedResponse(object):
    """
    Class wrapping a urllib3.HTTPResponse and representing a response whose
    body has not been yet fully read.

    It is a low-level representation used in the minet.web module that should
    not leak outside normally.

    It is able to stream a response's body safely all while keeping track of
    a "final" timeout correctly enforced to bypass python socket race condition
    issues on read loops and is also able to be cancelled if required.

    NOTE: this is the user's responsibility to close or unwrap the response
    after use. This will be done by __del__ in any case, but don't rely
    on it too much.
    """

    __slots__ = (
        "__inner",
        "__body",
        "__cancel_event",
        "__final_time",
        "__finished",
        "__closed",
    )

    def __init__(
        self,
        response: urllib3.HTTPResponse,
        cancel_event: Optional[Event],
        final_time: Optional[float] = None,
    ):
        self.__inner = response
        self.__cancel_event = cancel_event
        self.__final_time = final_time
        self.__body = BytesIO()
        self.__finished = False
        self.__closed = False

    def __len__(self) -> int:
        return self.__body.getbuffer().nbytes

    def __stream(
        self, chunk_size: int = STREAMING_CHUNK_SIZE, up_to: Optional[int] = None
    ) -> None:
        if self.__closed:
            raise TypeError("buffered response was already closed")

        if self.__finished:
            return

        self.__finished = stream_response_body(
            self.__inner,
            chunk_size=chunk_size,
            cancel_event=self.__cancel_event,
            final_time=self.__final_time,
            body=self.__body,
            up_to=up_to,
        )

    def geturl(self) -> Optional[str]:
        return self.__inner.geturl()

    def getheader(self, name: str, default: Optional[str] = None) -> Optional[str]:
        return self.__inner.getheader(name, default)

    @property
    def status(self) -> int:
        return self.__inner.status

    @property
    def body(self) -> bytes:
        return self.__body.getvalue()

    def close(self) -> None:
        if self.__closed:
            raise TypeError("already closed")

        self.__closed = True

        # urllib3's documentation is not very clear on the subject but it seems
        # the library was not geared toward only reading a few bytes from the
        # body of a request.
        # This means that if we read everything, we should only release the
        # connection so it can be reused by the pool. In the contrary,
        # we need to close the connection to avoid issues. What's more, it
        # is important to close the connection before releasing it.
        # Ref: https://urllib3.readthedocs.io/en/stable/advanced-usage.html#streaming-and-i-o
        if not self.__finished:
            # NOTE: if the response is 3xx, we can afford to drain the connection
            # not to lose it
            if self.__inner.status in REDIRECT_STATUSES:
                self.__inner.drain_conn()
            else:
                # NOTE: closing connections has a performance cost but I am
                # not really able to understand whether it would be safe not
                # to close them at all.
                self.__inner.close()

        self.__inner.release_conn()

    def __del__(self):
        if not self.__closed:
            # warnings.warn("BufferedResponse instance was not properly closed!")
            self.close()

    def unwrap(self) -> Tuple[urllib3.HTTPResponse, bytes]:
        self.close()
        return self.__inner, self.__body.getvalue()

    def read(self, chunk_size: int = STREAMING_CHUNK_SIZE) -> None:
        self.__stream(chunk_size=chunk_size)

    def read_and_unwrap(self) -> Tuple[urllib3.HTTPResponse, bytes]:
        self.read()
        return self.unwrap()

    def prebuffer_up_to(
        self, amount: int, chunk_size: int = STREAMING_CHUNK_SIZE
    ) -> None:
        self.__stream(chunk_size=chunk_size, up_to=amount)

    def extract_cookies_in_jar(self, jar: CookieJar) -> None:
        jar.extract_cookies(self.__inner, Request(self.geturl()))  # type: ignore


def atomic_request(
    pool_manager: urllib3.PoolManager,
    url: str,
    method="GET",
    headers=None,
    timeout: Optional[AnyTimeout] = None,
    body=None,
    cancel_event: Optional[Event] = None,
    final_time: Optional[float] = None,
) -> BufferedResponse:
    """
    Generic request helpers using a urllib3 pool_manager to access some resource.
    """

    # Validating URL
    if not ural.is_url(
        url, require_protocol=True, tld_aware=True, allow_spaces_in_path=True
    ):
        raise InvalidURLError(url)

    # We check for cancellation
    if cancel_event is not None and cancel_event.is_set():
        raise CancelledRequestError

    # Performing request
    request_kwargs = {
        "headers": headers,
        "body": body,
        "preload_content": False,
        "decode_content": True,
        "release_conn": False,
        "redirect": False,
        "retries": False,
    }

    if timeout is not None:
        request_kwargs["timeout"] = timeout

    if final_time is None:
        final_time = pool_manager_aware_timeout_to_final_time(pool_manager, timeout)

    response = pool_manager.request(method, url, **request_kwargs)

    return BufferedResponse(response, cancel_event=cancel_event, final_time=final_time)


def atomic_resolve(
    pool_manager: urllib3.PoolManager,
    url: str,
    method: str = "GET",
    headers=None,
    max_redirects: int = 5,
    follow_refresh_header: bool = True,
    follow_meta_refresh: bool = False,
    follow_js_relocation: bool = False,
    infer_redirection: bool = False,
    timeout: Optional[AnyTimeout] = None,
    cancel_event: Optional[Event] = None,
    final_time: Optional[float] = None,
    body=None,
    canonicalize: bool = False,
    stateful: bool = False,
) -> Tuple[RedirectionStack, BufferedResponse]:
    """
    Helper function attempting to resolve the given url.
    """

    url_stack: OrderedDict[str, Redirection] = OrderedDict()
    buffered_response: Optional[BufferedResponse] = None

    if final_time is None:
        final_time = pool_manager_aware_timeout_to_final_time(pool_manager, timeout)

    jar = None

    if stateful:
        jar = CookieJar()

    try:
        for _ in range(max_redirects):
            # We close last buffered_response as it won't be used anymore
            # NOTE: this should always happen at the beginning of the loop
            if buffered_response is not None:
                if stateful:
                    assert jar is not None
                    buffered_response.extract_cookies_in_jar(jar)

                buffered_response.close()
                # NOTE: resetting variable to avoid double free on errors
                buffered_response = None

            # Detecting cycles
            if not stateful and url in url_stack:
                raise InfiniteRedirectsError("Infinite redirects")

            if infer_redirection:
                target = ural.infer_redirection(url, recursive=False)

                if target != url:
                    url_stack[url] = Redirection(url, "infer")
                    url = target
                    continue

            redirection = Redirection(url)

            if stateful:
                assert jar is not None

                dummy_req = Request(url)
                jar.add_cookie_header(dummy_req)
                cookie = dummy_req.get_header("Cookie")

                if cookie is not None:
                    headers = {**(headers or {}), "Cookie": cookie}

            # NOTE: atomic_request will check for cancellation
            buffered_response = atomic_request(
                pool_manager,
                url,
                method=method,
                headers=headers,
                body=body,
                timeout=timeout,
                final_time=final_time,
                cancel_event=cancel_event,
            )

            redirection.status = buffered_response.status
            url_stack[url] = redirection

            # Attempting to find next location
            location = None

            if buffered_response.status not in REDIRECT_STATUSES:
                if buffered_response.status < 400:
                    # Refresh header
                    if follow_refresh_header:
                        refresh = buffered_response.getheader("refresh")

                        if refresh is not None:
                            p = parse_http_refresh(refresh)

                            if p is not None:
                                location = p[1]
                                redirection.type = "refresh-header"

                    # Reading a small chunk of the html
                    if location is None and (
                        follow_meta_refresh or follow_js_relocation
                    ):
                        buffered_response.prebuffer_up_to(
                            CONTENT_PREBUFFER_UP_TO
                            if not follow_js_relocation
                            else LARGE_CONTENT_PREBUFFER_UP_TO
                        )

                    # Meta refresh
                    if location is None and follow_meta_refresh:
                        meta_refresh = extract_meta_refresh(buffered_response.body)

                        if meta_refresh is not None:
                            location = meta_refresh[1]
                            redirection.type = "meta-refresh"

                    # JavaScript relocation
                    if location is None and follow_js_relocation:
                        js_relocation = extract_javascript_relocation(
                            buffered_response.body
                        )

                        if js_relocation is not None:
                            location = js_relocation
                            redirection.type = "js-relocation"

                # Found the end
                if location is None:
                    redirection.type = "hit"

                # Canonical url
                if redirection.type == "hit":
                    if canonicalize:
                        buffered_response.prebuffer_up_to(LARGE_CONTENT_PREBUFFER_UP_TO)

                        canonical = extract_canonical_link(buffered_response.body)

                        if canonical is not None and canonical != url:
                            try:
                                canonical = urljoin(url, canonical)
                            except ValueError:
                                raise InvalidRedirectError("Canonical url is invalid")

                            redirection = Redirection(canonical, "canonical")
                            url_stack[canonical] = redirection

                    # Breaking free of the retry loop because we have a hit
                    break

            else:
                redirection.type = "location-header"
                location = buffered_response.getheader("location")

            # Invalid redirection
            if not location:
                raise InvalidRedirectError("Redirection is invalid")

            # Location badly encoded?
            # NOTE: we don't really have a way to consume headers as raw bytes
            # which means we must rely on encoding duck-typing. I expect this
            # to fail in a complicated manner in a distant future.
            try:
                if not location.isascii():
                    byte_location = location.encode("latin1")
                    encoding = infer_encoding(byte_location)

                    if (
                        encoding is not None
                        and encoding != "iso-8859-1"
                        and encoding != "latin1"
                        and encoding != "ascii"
                    ):
                        location = byte_location.decode(encoding)
            except (UnicodeEncodeError, UnicodeDecodeError):
                raise BadlyEncodedLocationHeaderError(
                    "Location header has invalid encoding"
                )

            # Resolving next url
            try:
                next_url = urljoin(url, location.strip())
            except ValueError:
                # NOTE: sometimes this will fail because the next location is invalid
                raise InvalidRedirectError("Next location is an invalid url")

            # Self loop?
            if not stateful and next_url == url:
                raise SelfRedirectError("Self redirection")

            # Go to next
            url = next_url

        # We reached max redirects
        else:
            raise MaxRedirectsError("Maximum number of redirects exceeded")

    # NOTE: using BaseException here to avoid leaks on e.g. KeyboardInterrupt
    except BaseException:
        # NOTE: here we must make sure to cleanup any still alive
        # buffered response to avoid leaks
        if buffered_response is not None:
            buffered_response.close()

        raise

    # Compiling the stack and returning the last buffered response, still alive
    compiled_stack = list(url_stack.values())

    assert buffered_response is not None

    return compiled_stack, buffered_response


def build_request_headers(
    headers: Optional[Dict[str, str]] = None,
    cookie: Optional[Union[str, Dict[str, str]]] = None,
    spoof_ua: bool = False,
    json_body: bool = False,
    urlencoded_body: bool = False,
    compressed: bool = False,
):
    # Formatting headers
    final_headers = {}

    if spoof_ua:
        final_headers["User-Agent"] = DEFAULT_SPOOFED_UA

    if cookie:
        if not isinstance(cookie, str):
            cookie = dict_to_cookie_string(cookie)

        final_headers["Cookie"] = cookie

    if compressed:
        final_headers["Accept-Encoding"] = ACCEPT_ENCODING

    if json_body:
        final_headers["Content-Type"] = "application/json"
    elif urlencoded_body:
        final_headers["Content-Type"] = "application/x-www-form-urlencoded"

    # Note: headers passed explicitly by users always win
    if headers is not None:
        final_headers.update(headers)

    return final_headers


class Response(object):
    """
    Class representing a finalized HTTP response.

    It wraps a response binary body, headers & status and exposes a
    variety of useful utilities that will be used by end users.

    This class is used by the high-level functions of this module and is
    expected to be found elsewhere.

    Note that it will lazily compute required items when asked for certain
    properties.
    """

    __slots__ = (
        "__headers",
        "__status",
        "__stack",
        "__body",
        "__text",
        "__url",
        "__datetime_utc",
        "__is_text",
        "__encoding",
        "__ext",
        "__mimetype",
        "__has_guessed_extension",
        "__has_guessed_encoding",
        "__has_decoded_text",
    )

    __headers: HTTPHeaderDict
    __status: int
    __stack: Optional[RedirectionStack]
    __body: bytes
    __text: Optional[str]
    __url: str
    __datetime_utc: datetime
    __is_text: bool
    __encoding: Optional[str]
    __ext: Optional[str]
    __mimetype: Optional[str]

    __has_guessed_extension: bool
    __has_guessed_encoding: bool
    __has_decoded_text: bool

    def __init__(
        self,
        url: str,
        stack: Optional[RedirectionStack],
        headers: HTTPHeaderDict,
        status: int,
        body: bytes,
        known_encoding: Optional[str] = "utf-8",
    ):
        self.__url = url
        self.__stack = stack
        self.__headers = headers
        self.__status = status
        self.__body = body
        self.__text = None
        self.__datetime_utc = datetime.utcnow()
        self.__is_text = True
        self.__encoding = known_encoding
        self.__ext = None
        self.__mimetype = None

        self.__has_guessed_extension = False
        self.__has_guessed_encoding = known_encoding is not None
        self.__has_decoded_text = False

    def __guess_extension(self) -> None:
        if self.__has_guessed_extension:
            return

        # Guessing mime type
        # TODO: validate mime type string?
        url = self.__url
        assert url is not None
        mimetype, _ = mimetypes.guess_type(url)

        headers = self.__headers

        if "Content-Type" in headers:
            content_type = headers["Content-Type"]
            parsed_header = cgi.parse_header(content_type)

            if parsed_header and parsed_header[0].strip():
                mimetype = parsed_header[0].strip()

                # Dropping charset info
                if "," in mimetype:
                    mimetype = mimetype.split(",", 1)[0]

        if mimetype is None and looks_like_html(self.__body):
            mimetype = "text/html"

        if mimetype is not None:
            ext = mimetypes.guess_extension(mimetype)

            if ext == ".htm":
                ext = ".html"
            elif ext == ".jpe":
                ext = ".jpg"

            self.__mimetype = mimetype
            self.__ext = ext

        if mimetype is not None:
            self.__is_text = not is_binary_mimetype(mimetype)

        self.__has_guessed_extension = True

    def __guess_encoding(self) -> None:
        if self.__has_guessed_encoding:
            return

        self.__guess_extension()

        if not self.__is_text:
            return

        self.__encoding = infer_encoding(self.__body)

        self.__has_guessed_encoding = True

    def __decode(self, errors: str = "replace") -> None:
        if self.__has_decoded_text:
            return

        self.__guess_extension()
        self.__guess_encoding()

        if not self.__is_text:
            raise TypeError("response is binary and cannot be decoded")

        self.__text = self.__body.decode(self.__encoding or "utf-8", errors=errors)
        self.__has_decoded_text = True

    @property
    def url(self) -> str:
        return self.__url

    @property
    def start_url(self) -> str:
        return self.__url

    @property
    def end_url(self) -> str:
        if self.__stack is None:
            return self.__url
        return self.__stack[-1].url

    def urljoin(self, url: str) -> str:
        return urljoin(self.end_url, url)

    @property
    def headers(self):
        return self.__headers

    @property
    def stack(self) -> Optional[RedirectionStack]:
        return self.__stack

    @property
    def was_redirected(self) -> bool:
        return self.start_url != self.end_url

    @property
    def status(self) -> int:
        return self.__status

    @property
    def end_datetime(self) -> datetime:
        return self.__datetime_utc

    @property
    def ext(self) -> Optional[str]:
        self.__guess_extension()
        return self.__ext

    @property
    def mimetype(self) -> Optional[str]:
        self.__guess_extension()
        return self.__mimetype

    @property
    def is_text(self) -> bool:
        self.__guess_extension()
        return self.__is_text  # type: ignore # At that point we know it's a bool

    @property
    def is_binary(self) -> bool:
        return not self.is_text

    @property
    def could_be_html(self) -> bool:
        mime = self.mimetype
        return mime is not None and "/htm" in mime

    @property
    def is_html(self) -> bool:
        return self.is_text and looks_like_html(self.__body)

    @property
    def encoding(self) -> Optional[str]:
        self.__guess_encoding()
        return self.__encoding

    @property
    def likely_encoding(self) -> str:
        encoding = self.encoding

        if encoding is None:
            return "utf-8"

        return encoding

    @property
    def encoding_from_headers(self) -> Optional[str]:
        return get_encoding_from_content_type_header(self.__headers.get("Content-Type"))

    # TODO: add encoding_from_xml & possible_encodings when required

    @property
    def body(self) -> bytes:
        return self.__body

    def __len__(self) -> int:
        return len(self.__body)

    @property
    def human_size(self) -> str:
        return format_filesize(len(self.__body))

    def text(self) -> str:
        self.__decode()
        return self.__text  # type: ignore # If we don't raise, this IS a string

    def json(self):
        return json.loads(self.text())

    def soup(
        self,
        engine: str = "lxml",
        ignore_xhtml_warning=False,
        strainer: Optional[SoupStrainer] = None,
    ) -> WonderfulSoup:
        with suppress_xml_parsed_as_html_warnings(bypass=not ignore_xhtml_warning):
            return WonderfulSoup(self.text(), engine, parse_only=strainer)

    def links(self, unique: bool = True, strip_fragment: bool = False) -> List[str]:
        if not self.is_html:
            raise TypeError("cannot extract links from non-html responses")

        return list(
            ural.links_from_html(
                self.end_url,
                self.body,
                encoding=self.likely_encoding,
                canonicalize=True,
                unique=unique,
                strip_fragment=strip_fragment,
            )
        )

    def __repr__(self) -> str:
        attr: List[Union[str, Tuple[str, Union[str, bool]]]] = ["status", "url"]

        if self.url != self.end_url:
            attr.append(("resolved", self.end_url))

        attr.extend([("size", self.human_size), "mimetype"])

        if self.is_text:
            attr.append("encoding")

        return format_repr(self, attr)


def request(
    url: str,
    pool_manager: Optional[urllib3.PoolManager] = None,
    method: str = "GET",
    headers: Optional[Dict[str, str]] = None,
    cookie: Optional[Union[str, Dict[str, str]]] = None,
    spoof_ua: bool = False,
    follow_redirects: bool = True,
    max_redirects: int = 5,
    follow_refresh_header: bool = True,
    follow_meta_refresh: bool = False,
    follow_js_relocation: bool = False,
    infer_redirection: bool = False,
    canonicalize: bool = False,
    known_encoding: Optional[str] = "utf-8",
    timeout: Optional[AnyTimeout] = None,
    body: Optional[Union[str, bytes]] = None,
    json_body: Optional[Any] = None,
    urlencoded_body: Optional[Mapping[str, Union[str, int, float]]] = None,
    cancel_event: Optional[Event] = None,
    raise_on_statuses: Optional[Container[int]] = None,
    stateful: bool = False,
    use_pycurl: bool = False,
    compressed: bool = False,
) -> Response:
    # Pycurl and pool manager
    if use_pycurl:
        if not PYCURL_SUPPORT:
            raise PycurlNotInstalledError

        if pool_manager is not None:
            raise TypeError("use_pycurl is not compatible with pool_manager")
    elif pool_manager is None:
        pool_manager = DEFAULT_POOL_MANAGER

    # Formatting headers
    final_headers = build_request_headers(
        headers=headers,
        cookie=cookie,
        spoof_ua=spoof_ua,
        json_body=json_body is not None,
        compressed=compressed and not use_pycurl,
    )

    # Dealing with body
    if json_body is not None:
        body = json.dumps(json_body, ensure_ascii=False).encode("utf-8")
    elif urlencoded_body is not None:
        body = "&".join(
            "%s=%s" % (quote(str(k)), quote(str(v))) for k, v in urlencoded_body.items()
        )

    if isinstance(body, str):
        body = body.encode("utf-8")

    stack: Optional[RedirectionStack] = None

    if use_pycurl:
        if body is not None:
            raise NotImplementedError

        pycurl_result = request_with_pycurl(
            url,
            method=method,
            headers=final_headers,
            follow_redirects=follow_redirects,
            max_redirects=max_redirects,
            timeout=timeout,
            cancel_event=cancel_event,
            compressed=compressed,
        )

        if raise_on_statuses is not None and pycurl_result.status in raise_on_statuses:
            raise InvalidStatusError(pycurl_result.status)

        return Response(
            url,
            stack=pycurl_result.stack,
            headers=pycurl_result.headers,
            status=pycurl_result.status,
            body=pycurl_result.body,
            known_encoding=known_encoding,
        )

    assert pool_manager is not None

    if not follow_redirects:
        buffered_response = atomic_request(
            pool_manager,
            url,
            method,
            headers=final_headers,
            body=body,
            timeout=timeout,
            cancel_event=cancel_event,
        )
    else:
        stack, buffered_response = atomic_resolve(
            pool_manager,
            url,
            method,
            headers=final_headers,
            body=body,
            max_redirects=max_redirects,
            follow_refresh_header=follow_refresh_header,
            follow_meta_refresh=follow_meta_refresh,
            follow_js_relocation=follow_js_relocation,
            infer_redirection=infer_redirection,
            canonicalize=canonicalize,
            timeout=timeout,
            cancel_event=cancel_event,
            stateful=stateful,
        )

    if raise_on_statuses is not None and buffered_response.status in raise_on_statuses:
        buffered_response.close()
        raise InvalidStatusError(buffered_response.status)

    response, body = buffered_response.read_and_unwrap()

    return Response(
        url,
        stack=stack,
        headers=response.headers,
        status=response.status,
        body=body,
        known_encoding=known_encoding,
    )


def resolve(
    url: str,
    pool_manager: urllib3.PoolManager = DEFAULT_POOL_MANAGER,
    method: str = "GET",
    headers: Optional[Dict[str, str]] = None,
    cookie: Optional[Union[str, Dict[str, str]]] = None,
    spoof_ua: bool = False,
    max_redirects: int = 5,
    follow_refresh_header: bool = True,
    follow_meta_refresh: bool = False,
    follow_js_relocation: bool = False,
    infer_redirection: bool = False,
    timeout: Optional[AnyTimeout] = None,
    canonicalize: bool = False,
    cancel_event: Optional[Event] = None,
    raise_on_statuses: Optional[Container[int]] = None,
    stateful: bool = False,
) -> RedirectionStack:
    final_headers = build_request_headers(
        headers=headers, cookie=cookie, spoof_ua=spoof_ua
    )

    stack, buffered_response = atomic_resolve(
        pool_manager,
        url,
        method,
        headers=final_headers,
        max_redirects=max_redirects,
        follow_refresh_header=follow_refresh_header,
        follow_meta_refresh=follow_meta_refresh,
        follow_js_relocation=follow_js_relocation,
        infer_redirection=infer_redirection,
        timeout=timeout,
        canonicalize=canonicalize,
        cancel_event=cancel_event,
        stateful=stateful,
    )

    buffered_response.close()

    if raise_on_statuses is not None and stack:
        last = stack[-1]

        if last.status is not None and last.status in raise_on_statuses:
            raise InvalidStatusError(last.status)

    return stack


def request_jsonrpc(
    url: str,
    method: str,
    pool_manager: urllib3.PoolManager = DEFAULT_POOL_MANAGER,
    *args,
    **kwargs
) -> Response:
    params = []

    if len(args) > 0:
        params = args
    elif len(kwargs) > 0:
        params = kwargs

    return request(
        url,
        pool_manager=pool_manager,
        method="POST",
        json_body={"method": method, "params": params},
    )


ONE_DAY = 24 * 60 * 60


class log_request_retryer_before_sleep:
    def __init__(
        self, epilog_builder: Optional[Callable[[RetryCallState], Optional[str]]] = None
    ):
        if epilog_builder is not None and not callable(epilog_builder):
            raise TypeError("epilog_builder should be None or callable")

        self.epilog_builder = epilog_builder

    def __call__(self, retry_state: RetryCallState) -> None:
        assert retry_state.outcome is not None

        if not retry_state.outcome.failed:
            raise NotImplementedError

        exception = retry_state.outcome.exception()

        sleepers_logger.warn(
            "request_retryer starts sleeping",
            extra={
                "source": "request_retryer",
                "from_thread": threading.current_thread()
                is not threading.main_thread(),
                "retry_state": retry_state,
                "exception": exception,
                "sleep_time": retry_state.next_action.sleep,
                "epilog": self.epilog_builder(retry_state)
                if self.epilog_builder is not None
                else None,
            },
        )


class request_retryer_custom_exponential_backoff(wait_base):
    def __init__(self, min: float = 10, max: float = ONE_DAY, exp_base=4) -> None:
        self.min = min
        self.max = max
        self.exp_base = exp_base

    def __call__(self, retry_state: RetryCallState) -> float:
        # NOTE: we add/subtract randomly up to 1/4 of expected sleep time
        jitter = 0.5 - random.random()

        try:
            exp = self.exp_base ** (retry_state.attempt_number - 1)
            result = self.min * exp
            result += result / 2 * jitter
        except OverflowError:
            result = self.max

        return max(0, min(result, self.max))


class RequestRetrying(Retrying):
    def __init__(
        self, *args, invalid_statuses: Optional[Container[int]] = None, **kwargs
    ):
        self._invalid_statuses = invalid_statuses
        super().__init__(*args, **kwargs)

    def __call__(self, fn, *args, **kwargs):
        if self._invalid_statuses is not None and fn in (request, resolve):
            kwargs["raise_on_statuses"] = self._invalid_statuses

        return super().__call__(fn, *args, **kwargs)


def create_request_retryer(
    min: float = 10,
    max: float = ONE_DAY,
    max_attempts: int = 9,
    before_sleep=noop,
    additional_exceptions=None,
    retry_on_timeout: bool = True,
    retry_on_statuses: Optional[Container[int]] = None,
    predicate: Optional[Callable[[BaseException], bool]] = None,
    epilog: Optional[Callable[[RetryCallState], Optional[str]]] = None,
    cancel_event: Optional[Event] = None,
) -> RequestRetrying:
    # By default we only retry network issues, such as Internet being cut off etc.
    retryable_exception_types = [
        # urllib3 errors
        urllib3_exceptions.ProtocolError,
        # urllib errors
        urllib.error.URLError,
        # Low-level socket connection errors
        ConnectionAbortedError,
        ConnectionRefusedError,
        ConnectionResetError,
        # pycurl errors
        PycurlProtocolError,
    ]

    # We also usually include most timeout errors
    if retry_on_timeout:
        retryable_exception_types.extend(
            [FinalTimeoutError, urllib3_exceptions.TimeoutError, PycurlTimeoutError]
        )

    if retry_on_statuses is not None:
        retryable_exception_types.append(InvalidStatusError)

    if additional_exceptions:
        for exc in additional_exceptions:
            retryable_exception_types.append(exc)

    retryable_exception_types = tuple(retryable_exception_types)

    def default_retry_exception_predicate(exc: BaseException) -> bool:
        if not isinstance(exc, urllib3_exceptions.NewConnectionError):
            return isinstance(exc, retryable_exception_types)

        msg = str(exc).lower()

        if "errno -3" in msg or "temporary failure in name resolution" in msg:
            return True

        return False

    retry_condition = retry_if_exception(default_retry_exception_predicate)

    if retry_on_statuses is not None:

        def status_predicate(exc: BaseException) -> bool:
            if isinstance(exc, InvalidStatusError) and exc.status in retry_on_statuses:
                return True

            return False

        retry_condition |= retry_if_exception(status_predicate)

    if callable(predicate):
        retry_condition |= retry_if_exception(predicate)

    retrying_kwargs = {
        "reraise": True,
        "wait": request_retryer_custom_exponential_backoff(min=min, max=max),
        "retry": retry_condition,
        "stop": stop_after_attempt(max_attempts),
        "before_sleep": rcompose(
            log_request_retryer_before_sleep(epilog), before_sleep
        ),
    }

    # Cancellable sleep?
    if cancel_event is not None:
        if cancel_event.is_set():
            raise TypeError("cannot retry using an already set cancel_event")

        retrying_kwargs["sleep"] = sleep_using_event(cancel_event)
        retrying_kwargs["stop"] |= stop_when_event_set(cancel_event)
        retrying_kwargs["retry"] &= retry_if_exception(
            lambda _: not cancel_event.is_set()
        )

    return RequestRetrying(invalid_statuses=retry_on_statuses, **retrying_kwargs)


def retrying_method(attr="retryer"):
    def decorate(fn):
        @functools.wraps(fn)
        def decorated(self, *args, **kwargs):
            retryer = getattr(self, attr)

            if not isinstance(retryer, Retrying):
                raise ValueError

            return retryer(fn, self, *args, **kwargs)

        return decorated

    return decorate


class ThreadsafeRequestRetryers:
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.local_context = threading.local()

    def acquire(self) -> RequestRetrying:
        retryer = getattr(self.local_context, "retryer", None)

        if retryer is None:
            retryer = create_request_retryer(**self.kwargs)
            setattr(self.local_context, "retryer", retryer)

        return retryer


def threadsafe_retrying_method(attr="retryers"):
    def decorate(fn):
        @functools.wraps(fn)
        def decorated(self, *args, **kwargs):
            retryers = getattr(self, attr)

            if not isinstance(retryers, ThreadsafeRequestRetryers):
                raise ValueError

            retryer = retryers.acquire()

            return retryer(fn, self, *args, **kwargs)

        return decorated

    return decorate



# =============================================================================
# Minet CrowdTangle Links Summary CLI Action
# =============================================================================
#
# Logic of the `ct summary` action.
#
import casanova

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.crowdtangle.utils import with_crowdtangle_utilities
from minet.crowdtangle.types import CrowdTanglePost, CrowdTangleSummary

# TODO: could be a nested loading bar


@with_crowdtangle_utilities
@with_enricher_and_loading_bar(
    headers=CrowdTangleSummary, title="Collecting data", unit="urls"
)
def action(cli_args, client, enricher, loading_bar):
    posts_writer = None

    if cli_args.posts is not None:
        posts_writer = casanova.writer(
            cli_args.posts, fieldnames=["url"] + CrowdTanglePost.fieldnames()
        )

    for row, url in enricher.cells(cli_args.column, with_rows=True):
        url = url.strip()

        stats = client.summary(
            url,
            start_date=cli_args.start_date,
            with_top_posts=cli_args.posts is not None,
            sort_by=cli_args.sort_by,
            platforms=cli_args.platforms,
        )

        if cli_args.posts is not None:
            stats, posts = stats

            if posts is not None and posts_writer is not None:
                for post in posts:
                    posts_writer.writerow([url], post)

        enricher.writerow(row, stats.as_csv_row() if stats is not None else None)

        loading_bar.advance()



# =============================================================================
# Minet Twitter Scrape CLI Action
# =============================================================================
#
# Logic of the `tw scrape` action.
#
from twitwi.constants import USER_FIELDS
from twitwi import format_user_as_csv_row

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.exceptions import FatalError
from minet.twitter import TwitterAPIScraper
from minet.twitter.exceptions import (
    TwitterPublicAPIInvalidCookieError,
    TwitterPublicAPIBadAuthError,
)


@with_enricher_and_loading_bar(
    headers=USER_FIELDS,
    title="Scraping",
    unit="users",
    nested=True,
    sub_unit="followers",
)
def action(cli_args, enricher, loading_bar):
    try:
        scraper = TwitterAPIScraper(cli_args.cookie)
    except TwitterPublicAPIInvalidCookieError:
        raise FatalError(
            [
                "Invalid Twitter cookie!",
                "Try giving another browser to --cookie and sure you are correctly logged in.",
            ]
        )

    for row, user_id in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(user_id):
            iterator = scraper.followers_you_know(user_id, locale=cli_args.timezone)

            try:
                for user in iterator:
                    addendum = format_user_as_csv_row(user)
                    enricher.writerow(row, addendum)
                    loading_bar.nested_advance()

            except TwitterPublicAPIBadAuthError as error:
                raise FatalError(
                    "Bad authentication (%i). Double check your --cookie and make sure you are logged in."
                    % error.status
                )



# =============================================================================
# Minet Scraper Analysis
# =============================================================================
#
# Functions performing analysis of scraper definitions.
#
from typing import Optional, List
from minet.types import Literal

import ast
import soupsieve
from soupsieve import SelectorSyntaxError

from minet.scrape.utils import get_sel, get_iterator
from minet.scrape.constants import (
    PLURAL_MODIFIERS,
    BURROWING_KEYS,
    LEAF_KEYS,
    EXTRACTOR_NAMES,
    KNOWN_KEYS,
)
from minet.scrape.exceptions import (
    ScraperEvalSyntaxError,
    ScraperValidationConflictError,
    InvalidCSSSelectorError,
    ScraperValidationIrrelevantPluralModifierError,
    ScraperValidationMixedConcernError,
    ScraperValidationInvalidPluralModifierError,
    ScraperValidationInvalidExtractorError,
    ScraperValidationUnknownKeyError,
)

ScraperAnalysisOutputType = Literal["scalar", "unknown", "collection", "dict", "list"]


class ScraperAnalysis(object):
    fieldnames: Optional[List[str]]
    plural: bool
    output_type: ScraperAnalysisOutputType

    __slots__ = ("fieldnames", "plural", "output_type")

    def __init__(
        self,
        fieldnames: Optional[List[str]] = None,
        plural: bool = False,
        output_type: ScraperAnalysisOutputType = "scalar",
    ):
        self.fieldnames = fieldnames
        self.plural = plural
        self.output_type = output_type

    def __eq__(self, other):
        return (
            self.plural == other.plural
            and self.output_type == other.output_type
            and (
                (self.fieldnames is None and other.fieldnames is None)
                or (set(self.fieldnames) == set(other.fieldnames))
            )
        )

    def __repr__(self):
        return "<{name} fieldnames={fieldnames!r} plural={plural} output_type={output_type}>".format(
            name=self.__class__.__name__,
            fieldnames=self.fieldnames,
            plural=self.plural,
            output_type=self.output_type,
        )


def fieldnames_from_definition(scraper):
    tabulate = scraper.get("tabulate")

    if tabulate is not None:
        if not isinstance(tabulate, dict):
            return None

        return tabulate.get("headers")

    fields = scraper.get("fields")

    if fields is None:
        return ["value"]
    else:
        for node in fields.values():
            if "fields" in node:
                return None

    return list(fields.keys())


def analyse(scraper):
    analysis = ScraperAnalysis()

    analysis.fieldnames = fieldnames_from_definition(scraper)

    iterator = get_iterator(scraper)

    if iterator is not None or "iterator_eval" in scraper:
        analysis.plural = True
        analysis.output_type = "list"

    if "fields" in scraper:
        if analysis.plural:
            analysis.output_type = "collection"
        else:
            analysis.output_type = "dict"

    elif "eval" in scraper:
        analysis.output_type = "unknown"

    return analysis


ERRORS_PRIORITY = {
    InvalidCSSSelectorError: 0,
    ScraperValidationUnknownKeyError: 1,
    ScraperEvalSyntaxError: 2,
    ScraperValidationConflictError: 3,
    ScraperValidationMixedConcernError: 4,
    ScraperValidationIrrelevantPluralModifierError: 5,
    ScraperValidationInvalidPluralModifierError: 6,
    ScraperValidationInvalidExtractorError: 7,
}


def errors_sorting_key(error):
    return (tuple(error.path), ERRORS_PRIORITY[type(error)])


def validate(scraper):
    errors = []

    def recurse(node, path=[]):
        # Checking conflicts
        c = []

        if "item" in node:
            c.append("item")
        if "fields" in node:
            c.append("fields")
        if "tabulate" in node:
            c.append("tabulate")

        if len(c) > 1:
            validation_error = ScraperValidationConflictError(path=path, keys=c)

            errors.append(validation_error)

        # Validating extractors
        if "extract" in node and node["extract"] not in EXTRACTOR_NAMES:
            errors.append(
                ScraperValidationInvalidExtractorError(
                    path=path, extractor=node["extract"]
                )
            )

        # Validating selectors
        # NOTE: this has the beneficial side effect of precompiling selectors
        k, sel = get_sel(node, with_key=True)

        if sel is not None:
            try:
                soupsieve.compile(sel)
            except (SelectorSyntaxError, NotImplementedError) as e:
                errors.append(
                    InvalidCSSSelectorError(path=path + [k], reason=e, expression=sel)
                )

        k, iterator = get_iterator(node, with_key=True)

        if iterator is not None:
            try:
                soupsieve.compile(iterator)
            except (SelectorSyntaxError, NotImplementedError) as e:
                errors.append(
                    InvalidCSSSelectorError(
                        path=path + [k], reason=e, expression=iterator
                    )
                )

        # Checking plural modifiers
        if iterator is None and "iterator_eval" not in node:
            for modifier in PLURAL_MODIFIERS:
                if modifier in node:
                    errors.append(
                        ScraperValidationIrrelevantPluralModifierError(
                            path=path, modifier=modifier
                        )
                    )

        for modifier in PLURAL_MODIFIERS:
            if modifier not in node or "_eval" in modifier:
                continue

            if isinstance(node[modifier], bool):
                if "fields" in node:
                    errors.append(
                        ScraperValidationInvalidPluralModifierError(
                            path=path, modifier=modifier
                        )
                    )
            else:
                if "fields" not in node:
                    errors.append(
                        ScraperValidationInvalidPluralModifierError(
                            path=path, modifier=modifier
                        )
                    )

        # Conflicting leaf keys
        conflicting_leaf_keys = []

        if "attr" in node:
            conflicting_leaf_keys.append("attr")
        if "extract" in node:
            conflicting_leaf_keys.append("extract")
        if "get_context" in node:
            conflicting_leaf_keys.append("get_context")

        if len(conflicting_leaf_keys) > 1:
            errors.append(
                ScraperValidationConflictError(path=path, keys=conflicting_leaf_keys)
            )

        # Conflicting burrowing/leaf
        if any(k in node for k in BURROWING_KEYS) and any(k in node for k in LEAF_KEYS):
            errors.append(ScraperValidationMixedConcernError(path=path))

        in_fields = path and path[-1] == "fields"

        for k, v in node.items():
            # Unknown keys
            if not in_fields and k not in KNOWN_KEYS:
                errors.append(ScraperValidationUnknownKeyError(path=path, key=k))
                continue

            p = path + [k]

            # Validating python syntax
            if k == "eval" or k.endswith("_eval"):
                if isinstance(v, str):
                    try:
                        ast.parse(v)
                    except SyntaxError as e:
                        validation_error = ScraperEvalSyntaxError(
                            reason=e, expression=v, path=p
                        )

                        errors.append(validation_error)
            else:
                k_eval = "%s_eval" % k

                if k_eval in node:
                    validation_error = ScraperValidationConflictError(
                        path=path, keys=[k, k_eval]
                    )

                    errors.append(validation_error)

            if isinstance(v, dict):
                recurse(v, p)

    recurse(scraper)

    return sorted(errors, key=errors_sorting_key)



# from: https://github.com/python/cpython/blob/3.8/Lib/asyncio/unix_events.py
import os
import itertools
import threading
import warnings
from asyncio import AbstractChildWatcher, get_running_loop
from asyncio.log import logger


def _compute_returncode(status):
    if os.WIFSIGNALED(status):
        # The child process died because of a signal.
        return -os.WTERMSIG(status)
    elif os.WIFEXITED(status):
        # The child process exited (e.g sys.exit()).
        return os.WEXITSTATUS(status)
    else:
        # The child exited, but we don't understand its status.
        # This shouldn't happen, but if it does, let's just
        # return that status; perhaps that helps debug it.
        return status


class ThreadedChildWatcher(AbstractChildWatcher):
    """Threaded child watcher implementation.

    The watcher uses a thread per process
    for waiting for the process finish.

    It doesn't require subscription on POSIX signal
    but a thread creation is not free.

    The watcher has O(1) complexity, its performance doesn't depend
    on amount of spawn processes.
    """

    def __init__(self):
        self._pid_counter = itertools.count(0)
        self._threads = {}

    def is_active(self):
        return True

    def close(self):
        self._join_threads()

    def _join_threads(self):
        """Internal: Join all non-daemon threads"""
        threads = [
            thread
            for thread in list(self._threads.values())
            if thread.is_alive() and not thread.daemon
        ]
        for thread in threads:
            thread.join()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __del__(self, _warn=warnings.warn):
        threads = [
            thread for thread in list(self._threads.values()) if thread.is_alive()
        ]
        if threads:
            _warn(
                f"{self.__class__} has registered but not finished child processes",
                ResourceWarning,
                source=self,
            )

    def add_child_handler(self, pid, callback, *args):
        loop = get_running_loop()
        thread = threading.Thread(
            target=self._do_waitpid,
            name=f"waitpid-{next(self._pid_counter)}",
            args=(loop, pid, callback, args),
            daemon=True,
        )
        self._threads[pid] = thread
        thread.start()

    def remove_child_handler(self, pid):
        # asyncio never calls remove_child_handler() !!!
        # The method is no-op but is implemented because
        # abstract base class requires it
        return True

    def attach_loop(self, loop):
        pass

    def _do_waitpid(self, loop, expected_pid, callback, args):
        assert expected_pid > 0

        try:
            pid, status = os.waitpid(expected_pid, 0)
        except ChildProcessError:
            # The child process is already reaped
            # (may happen if waitpid() is called elsewhere).
            pid = expected_pid
            returncode = 255
            logger.warning(
                "Unknown child process pid %d, will report returncode 255", pid
            )
        else:
            returncode = _compute_returncode(status)
            if loop.get_debug():
                logger.debug(
                    "process %s exited with returncode %s", expected_pid, returncode
                )

        if loop.is_closed():
            logger.warning("Loop %r that handles pid %r is closed", loop, pid)
        else:
            loop.call_soon_threadsafe(callback, pid, returncode, *args)

        self._threads.pop(expected_pid)



# =============================================================================
# Minet Facebook CLI Action
# =============================================================================
#
# Logic of the `fb` action.
#
from minet.cli.argparse import command, ConfigAction

# TODO: lazyloading issue
from minet.facebook.constants import FACEBOOK_MOBILE_DEFAULT_THROTTLE

MOBILE_ARGUMENTS = [
    {
        "flags": ["-c", "--cookie"],
        "help": 'Authenticated cookie to use or browser from which to extract it (supports "firefox", "chrome", "chromium", "opera" and "edge").',
        "default": "firefox",
        "rc_key": ["facebook", "cookie"],
        "action": ConfigAction,
    },
    {
        "flag": "--throttle",
        "help": "Throttling time, in seconds, to wait between each request.",
        "type": float,
        "default": FACEBOOK_MOBILE_DEFAULT_THROTTLE,
    },
]

FACEBOOK_COMMENTS_SUBCOMMAND = command(
    "comments",
    "minet.cli.facebook.comments",
    title="Minet Facebook Comments Command",
    description="""
        Scrape a Facebook post's comments.

        This requires to be logged in to a Facebook account, so
        by default this command will attempt to grab the relevant
        authentication cookies from a local Firefox browser.

        If you want to grab cookies from another browser or want
        to directly pass the cookie as a string, check out the
        -c/--cookie flag.
    """,
    epilog="""
        Examples:

        . Scraping a post's comments:
            $ minet fb comments https://www.facebook.com/groups/186982532676569/permalink/4096995827030341/ > comments.csv

        . Grabbing cookies from chrome:
            $ minet fb comments -c chrome https://www.facebook.com/groups/186982532676569/permalink/4096995827030341/ > comments.csv

        . Scraping comments from multiple posts listed in a CSV file:
            $ minet fb comments post_url -i posts.csv > comments.csv
    """,
    variadic_input={"dummy_column": "post_url", "item_label": "post url"},
    arguments=[*MOBILE_ARGUMENTS],
)

FACEBOOK_EXPERIMENTAL_COMMENTS_SUBCOMMAND = command(
    "experimental-comments",
    "minet.cli.facebook.experimental_comments",
    title="Minet Experimental Facebook Comments Command",
    description="""
        Scrape a Facebook post's comments using an experimental scraper
        based on browser emulation.
    """,
    epilog="""
        Examples:

        . Scraping a post's comments:
            $ minet fb experimental-comments https://www.facebook.com/groups/186982532676569/permalink/4096995827030341/ > comments.csv

        . Scraping comments from multiple posts listed in a CSV file:s
            $ minet fb experimental-comments post_url -i posts.csv > comments.csv
    """,
    variadic_input={"dummy_column": "post_url", "item_label": "post url"},
    arguments=[
        {
            "flag": "--headful",
            "help": "Whether to disable the browser's headless mode for debugging purposes.",
            "action": "store_true",
        }
    ],
)

FACEBOOK_POST_AUTHORS_SUBCOMMAND = command(
    "post-authors",
    "minet.cli.facebook.post_authors",
    title="Minet Facebook Post Authors Command",
    description="""
        Retrieve the author of the given Facebook posts.

        Note that it is only relevant for group posts since
        only administrators can post something on pages.
    """,
    epilog="""
        Examples:

        . Fetching authors of a series of posts in a CSV file:
            $ minet fb post-authors post_url -i fb-posts.csv > authors.csv
    """,
    variadic_input={"dummy_column": "post_url", "item_label": "post"},
    arguments=[*MOBILE_ARGUMENTS],
)

FACEBOOK_POST_STATS_SUBCOMMAND = command(
    "post-stats",
    "minet.cli.facebook.post_stats",
    title="Minet Facebook Post Stats Command",
    description="""
        Retrieve statistics about a given list of Facebook posts.
    """,
    epilog="""
        Examples:

        . Fetching stats about lists of posts in a CSV file:
            $ minet fb post-stats post_url -i fb-posts.csv > stats.csv
    """,
    variadic_input={"dummy_column": "post_url", "item_label": "post url"},
)

FACEBOOK_POST_SUBCOMMAND = command(
    "post",
    "minet.cli.facebook.post",
    title="Minet Facebook Post Command",
    description="""
        Scrape Facebook post.

        This requires to be logged in to a Facebook account, so
        by default this command will attempt to grab the relevant
        authentication cookies from a local Firefox browser.

        If you want to grab cookies from another browser or want
        to directly pass the cookie as a string, check out the
        -c/--cookie flag.

        You must set your account language to English (US) for the
        command to work.

        Note that, by default, Facebook will translate post text
        when they are not written in a language whitelisted here:
        https://www.facebook.com/settings/?tab=language

        In this case, minet will output both the original text and
        the translated one. But be aware that original text may be
        truncated, so you might want to edit your Facebook settings
        using the url above to make sure text won't be translated
        for posts you are interested in.

        Of course, the CLI will warn you when translated text is
        found so you can choose to edit your settings early as
        as possible.

        Finally, some post text is always truncated on Facebook
        when displayed in lists. This text is not yet entirely
        scraped by minet at this time.
    """,
    epilog="""
        Examples:

        . Scraping a post:
            $ minet fb post https://m.facebook.com/watch/?v=448540820705115 > post.csv

        . Grabbing cookies from chrome:
            $ minet fb posts -c chrome https://m.facebook.com/watch/?v=448540820705115 > post.csv

        . Scraping post from multiple urls listed in a CSV file:
            $ minet fb post url -i urls.csv > post.csv
    """,
    variadic_input={"dummy_column": "post_url", "item_label": "post url"},
    arguments=[*MOBILE_ARGUMENTS],
)

FACEBOOK_POSTS_SUBCOMMAND = command(
    "posts",
    "minet.cli.facebook.posts",
    title="Minet Facebook Posts Command",
    description="""
        Scrape Facebook posts.

        This requires to be logged in to a Facebook account, so
        by default this command will attempt to grab the relevant
        authentication cookies from a local Firefox browser.

        If you want to grab cookies from another browser or want
        to directly pass the cookie as a string, check out the
        -c/--cookie flag.

        Scraping posts currently only works for Facebook groups.

        Note that, by default, Facebook will translate post text
        when they are not written in a language whitelisted here:
        https://www.facebook.com/settings/?tab=language

        In this case, minet will output both the original text and
        the translated one. But be aware that original text may be
        truncated, so you might want to edit your Facebook settings
        using the url above to make sure text won't be translated
        for posts you are interested in.

        Of course, the CLI will warn you when translated text is
        found so you can choose to edit your settings early as
        as possible.

        Finally, some post text is always truncated on Facebook
        when displayed in lists. This text is not yet entirely
        scraped by minet at this time.
    """,
    epilog="""
        Examples:

        . Scraping a group's posts:
            $ minet fb posts https://www.facebook.com/groups/444175323127747 > posts.csv

        . Grabbing cookies from chrome:
            $ minet fb posts -c chrome https://www.facebook.com/groups/444175323127747 > posts.csv

        . Scraping posts from multiple groups listed in a CSV file:
            $ minet fb posts group_url -i groups.csv > posts.csv
    """,
    variadic_input={"dummy_column": "group_url", "item_label": "group url"},
    arguments=[*MOBILE_ARGUMENTS],
)

FACEBOOK_URL_LIKES_SUBCOMMAND = command(
    "url-likes",
    "minet.cli.facebook.url_likes",
    title="Minet Facebook Url Likes Command",
    description="""
        Retrieve the approximate number of "likes" (actually an aggregated engagement metric)
        that a url got on Facebook. The command can also be used with a list of urls stored in a CSV file.
        This number is found by scraping Facebook's share button, which only gives a
        rough estimation of the real engagement metric: "Share 45K" for example.

        Note that this number does not actually only correspond to the number of
        likes or shares, but it is rather the sum of like, love, ahah, angry, etc.
        reactions plus the number of comments and shares that the URL got on Facebook
        (here is the official documentation: https://developers.facebook.com/docs/plugins/faqs
        explaining "What makes up the number shown next to my Share button?").
    """,
    epilog="""
        example:
        . Retrieving the "like" number for one url:
            $ minet fb url-likes "www.example-url.com" > url_like.csv

        . Retrieving the "like" number for the urls listed in a CSV file:
            $ minet fb url-likes url -i url.csv > url_likes.csv
    """,
    variadic_input={"dummy_column": "url", "item_label": "url"},
)

FACEBOOK_COMMAND = command(
    "facebook",
    "minet.cli.facebook",
    "Minet Facebook Command",
    aliases=["fb"],
    description="""
        Collect data from Facebook.
    """,
    subcommands=[
        FACEBOOK_COMMENTS_SUBCOMMAND,
        FACEBOOK_EXPERIMENTAL_COMMENTS_SUBCOMMAND,
        FACEBOOK_POST_AUTHORS_SUBCOMMAND,
        FACEBOOK_POST_STATS_SUBCOMMAND,
        FACEBOOK_POST_SUBCOMMAND,
        FACEBOOK_POSTS_SUBCOMMAND,
        FACEBOOK_URL_LIKES_SUBCOMMAND,
    ],
)



# =============================================================================
# Minet CrowdTangle Lists CLI Action
# =============================================================================
#
# Logic of the `ct lists` action.
#
import casanova

from minet.cli.crowdtangle.utils import with_crowdtangle_utilities
from minet.crowdtangle.types import CrowdTangleList


@with_crowdtangle_utilities
def action(cli_args, client):
    writer = casanova.writer(cli_args.output, fieldnames=CrowdTangleList)

    lists = client.lists()

    for l in lists:
        writer.writerow(l)



import logging
from logging import NullHandler

main_logger = logging.getLogger("minet")
sleepers_logger = main_logger.getChild("sleepers")
downloaders_logger = main_logger.getChild("downloaders")

main_logger.addHandler(NullHandler())



# =============================================================================
# Minet Instagram Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class InstagramError(MinetError):
    pass


class InstagramPublicAPIInvalidResponseError(InstagramError):
    def __init__(self, url, status, data):
        super().__init__()
        self.url = url
        self.status = status
        self.data = data

    def __str__(self):
        return super().__str__() + ", Url: %s, Status: %i, Data: %s" % (
            self.url,
            self.status,
            self.data,
        )


class InstagramInvalidCookieError(InstagramError):
    pass


class InstagramTooManyRequestsError(InstagramError):
    pass


class InstagramError500(InstagramError):
    pass


class InstagramInvalidTargetError(InstagramError):
    pass


class InstagramNoPublicationError(InstagramError):
    pass


class InstagramPrivateOrNonExistentAccountError(InstagramError):
    pass


class InstagramHashtagNeverUsedError(InstagramError):
    pass


class InstagramPrivateAccountError(InstagramError):
    pass


class InstagramAccountNoFollowError(InstagramError):
    pass



from typing import Optional, List, Any, Dict, Type, cast

from bs4 import SoupStrainer, BeautifulSoup
from urllib.parse import urljoin
from ural import should_follow_href, could_be_rss

from minet.scrape.analysis import ScraperAnalysisOutputType
from minet.scrape.utils import ensure_soup
from minet.scrape.types import AnyScrapableTarget
from minet.scrape.classes.base import ScraperBase


class NamedScraper(ScraperBase):
    name: str
    fieldnames: List[str]
    plural: bool
    tabular = True
    output_type: ScraperAnalysisOutputType
    strainer: Optional[SoupStrainer]

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        raise NotImplementedError

    def __call__(self, html: AnyScrapableTarget, context=None) -> Any:
        soup = ensure_soup(html, strainer=self.strainer)
        return self.scrape(soup, context=context)


class TitleScraper(NamedScraper):
    name = "title"
    fieldnames = ["title"]
    plural = False
    output_type = "scalar"
    strainer = SoupStrainer(name="title")

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        title_elem = soup.find(name="title")

        if title_elem is None:
            return None

        return title_elem.get_text().strip()


class CanonicalScraper(NamedScraper):
    name = "canonical"
    fieldnames = ["canonical_url"]
    plural = False
    output_type = "scalar"
    strainer = SoupStrainer(name="link", attrs={"rel": "canonical"})

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        link_elem = soup.select_one("link[rel=canonical][href]")

        if link_elem is None:
            return None

        url = link_elem.get("href")

        if url is None:
            return None

        url = cast(str, url).strip()

        if not url:
            return None

        return url


class UrlsScraper(NamedScraper):
    name = "urls"
    fieldnames = ["url"]
    plural = True
    output_type = "list"
    strainer = SoupStrainer(name="a")

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        a_elems = soup.select("a[href]")
        base_url = context.get("url") if context is not None else None

        urls = []

        for a in a_elems:
            url = a.get("href")

            if url is None:
                continue

            url = cast(str, url).strip()

            if not url:
                continue

            if not should_follow_href(url):
                continue

            if base_url:
                url = urljoin(base_url, url)

            urls.append(url)

        return urls


class ImagesScraper(NamedScraper):
    name = "images"
    fieldnames = ["src"]
    plural = True
    output_type = "list"
    strainer = SoupStrainer(name="img")

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        img_elems = soup.select("img[src]")
        base_url = context.get("url") if context is not None else None

        urls = []

        for img in img_elems:
            url = img.get("src")

            if url is None:
                continue

            url = cast(str, url).strip()

            if not url:
                continue

            if not should_follow_href(url):
                continue

            if base_url:
                url = urljoin(base_url, url)

            urls.append(url)

        return urls


class MetasScraper(NamedScraper):
    name = "metas"
    fieldnames = [
        "name",
        "property",
        "http-equiv",
        "itemprop",
        "content",
        "charset",
    ]
    plural = True
    output_type = "collection"
    strainer = SoupStrainer(name="meta")

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        meta_elems = soup.find_all(name="meta")

        metas = []

        for meta_elem in meta_elems:
            metas.append({name: meta_elem.get(name) for name in self.fieldnames})

        return metas


class RssScraper(NamedScraper):
    name = "rss"
    fieldnames = ["rss_url"]
    plural = True
    output_type = "list"
    strainer = SoupStrainer(name=["a", "link"])

    def scrape(self, soup: BeautifulSoup, context=None):
        rss_urls = []
        base_url = context.get("url") if context is not None else ""

        for link in soup.find_all():
            if link.name == "link":
                type_attr = link.attrs.get("type", None)
                if (
                    type_attr == "application/rss+xml"
                    or type_attr == "application/atom+xml"
                ):
                    href = link.attrs.get("href", None)
                    if href:
                        rss_urls.append(urljoin(base_url, href))
            else:
                href = link.attrs.get("href", None)
                url = urljoin(base_url, href)
                if could_be_rss(url):
                    rss_urls.append(url)

        return rss_urls


NAMED_SCRAPERS: Dict[str, Type[NamedScraper]] = {
    s.name: s
    for s in [
        TitleScraper,
        CanonicalScraper,
        UrlsScraper,
        ImagesScraper,
        MetasScraper,
        RssScraper,
    ]
}



import calendar
import dateparser
from datetime import datetime


def timestamp_to_isoformat(timestamp):
    return datetime.utcfromtimestamp(timestamp).isoformat()


def parse_date(formatted_date, lang="en"):
    if not isinstance(formatted_date, str):
        raise TypeError

    parsed = dateparser.parse(formatted_date, languages=[lang])

    if not parsed:
        return None

    return parsed.isoformat().split(".", 1)[0]


PARTIAL_ISO_FORMATS = {
    4: (r"%Y", "year"),
    7: (r"%Y-%m", "month"),
    10: (r"%Y-%m-%d", "day"),
    13: (r"%Y-%m-%dT%H", "hour"),
    16: (r"%Y-%m-%dT%H:%M", "minute"),
    19: (r"%Y-%m-%dT%H:%M:%S", "second"),
    20: (r"%Y-%m-%dT%H:%M:%SZ", "second"),
}


def datetime_from_partial_iso_format(
    string: str, upper_bound: bool = False
) -> datetime:
    try:
        possible_date_format, precision = PARTIAL_ISO_FORMATS[len(string)]
    except KeyError:
        raise ValueError("cannot parse date {!r}".format(string))

    result = datetime.strptime(string, possible_date_format)

    if upper_bound:
        if precision == "year":
            result = result.replace(month=12, day=31, hour=23, minute=59, second=59)
        elif precision == "month":
            _, last_day = calendar.monthrange(result.year, result.month)
            result = result.replace(day=last_day, hour=23, minute=59, second=59)
        elif precision == "day":
            result = result.replace(hour=23, minute=59, second=59)
        elif precision == "hour":
            result = result.replace(minute=59, second=59)
        elif precision == "minute":
            result = result.replace(second=59)

    return result



from typing import Optional, List

from casanova import TabularRecord
from dataclasses import dataclass
from ebbe import getpath


def get_int(item, key) -> Optional[int]:
    nb = item.get(key)

    if nb is not None:
        return int(nb)

    return None


@dataclass
class YouTubeCaptionTrack(TabularRecord):
    lang: str
    url: str
    generated: bool


@dataclass
class YouTubeCaptionLine(TabularRecord):
    start: float
    duration: float
    text: str


@dataclass
class YouTubeVideoSnippet(TabularRecord):
    video_id: str
    title: str
    published_at: str
    description: str
    channel_id: str
    channel_title: str

    @classmethod
    def from_payload(cls, payload) -> "YouTubeVideoSnippet":
        snippet = payload["snippet"]

        return cls(
            video_id=payload["id"]["videoId"],
            published_at=snippet["publishedAt"],
            channel_id=snippet["channelId"],
            channel_title=snippet["title"],
            description=snippet["description"],
            title=snippet["channelTitle"],
        )


@dataclass
class YouTubePlaylistVideoSnippet(YouTubeVideoSnippet):
    position: int

    @classmethod
    def from_payload(cls, payload) -> "YouTubePlaylistVideoSnippet":
        snippet = payload["snippet"]

        return cls(
            video_id=snippet["resourceId"]["videoId"],
            published_at=snippet["publishedAt"],
            channel_id=snippet["channelId"],
            title=snippet["title"],
            description=snippet["description"],
            channel_title=snippet["channelTitle"],
            position=snippet["position"],
        )


@dataclass
class YouTubeVideo(YouTubeVideoSnippet):
    view_count: Optional[int]
    like_count: Optional[int]
    # dislike_count: int # NOTE: This property is deprecated since december 13th 2021.
    # favorite_count: int # NOTE: This property has been deprecated by YouTube in 2015. The property's value is now always set to 0.
    comment_count: Optional[int]
    duration: str
    has_captions: bool

    @classmethod
    def from_payload(cls, payload) -> "YouTubeVideo":
        snippet = payload["snippet"]
        stats = payload["statistics"]
        details = payload["contentDetails"]

        return cls(
            video_id=payload["id"],
            published_at=snippet["publishedAt"],
            channel_id=snippet["channelId"],
            title=snippet["title"],
            description=snippet["description"],
            channel_title=snippet["channelTitle"],
            view_count=get_int(stats, "viewCount"),
            like_count=get_int(stats, "likeCount"),
            comment_count=get_int(stats, "commentCount"),
            duration=details["duration"],
            has_captions=details["caption"] == "true",
        )


@dataclass
class YouTubeComment(TabularRecord):
    video_id: str
    comment_id: str
    author_name: str
    author_channel_id: str
    text: str
    like_count: int
    published_at: str
    updated_at: str
    reply_count: Optional[int]
    parent_comment_id: Optional[str]

    @classmethod
    def from_parent_comment_payload(cls, payload) -> "YouTubeComment":
        meta = payload["snippet"]
        snippet = getpath(payload, ["snippet", "topLevelComment", "snippet"])

        return cls(
            meta["videoId"],
            payload["id"],
            snippet["authorDisplayName"],
            getpath(snippet, ["authorChannelId", "value"]),
            snippet["textOriginal"],
            int(snippet["likeCount"]),
            snippet["publishedAt"],
            snippet["updatedAt"],
            int(meta["totalReplyCount"]),
            None,
        )

    @classmethod
    def from_reply_payload(
        cls, payload, video_id: Optional[str] = None
    ) -> "YouTubeComment":
        snippet = payload["snippet"]

        return cls(
            video_id if video_id is not None else snippet["videoId"],
            payload["id"],
            snippet["authorDisplayName"],
            getpath(snippet, ["authorChannelId", "value"]),
            snippet["textOriginal"],
            int(snippet["likeCount"]),
            snippet["publishedAt"],
            snippet["updatedAt"],
            None,
            snippet["parentId"],
        )


@dataclass
class YouTubeChannel(TabularRecord):
    channel_id: str
    title: str
    description: str
    custom_url: str
    published_at: str
    thumbnail: str
    default_language: str
    country: str
    id_playlists_videos: List[str]
    view_count: int
    hidden_subscriber_count: int
    subscriber_count: int
    video_count: int
    topic_ids: List[str]
    topic_categories: List[str]
    topic_keywords: List[str]
    privacy_status: str
    made_for_kids: bool
    long_uploads_status: str
    keywords: List[str]
    moderate_comments: bool
    unsubscribed_trailer: str
    banner_external_url: str

    @classmethod
    def from_payload(cls, payload) -> "YouTubeChannel":
        snippet = payload.get("snippet")
        statistics = payload.get("statistics")
        topic_details = payload.get("topicDetails", {})
        status = payload.get("status")
        branding_settings = payload.get("brandingSettings")

        topic_ids = topic_details.get("topicIds", [])
        topic_categories = topic_details.get("topicCategories", [])
        topic_keywords = [url.rsplit("/", 1)[1] for url in topic_categories]

        keywords = getpath(branding_settings, ["channel", "keywords"])
        if keywords:
            keywords = [
                keyword.strip() for keyword in keywords.split('"') if keyword.strip()
            ]

        return YouTubeChannel(
            payload.get("id"),
            snippet.get("title"),
            snippet.get("description"),
            snippet.get("customUrl"),
            snippet.get("publishedAt"),
            getpath(snippet, ["thumbnails", "high", "url"]),
            snippet.get("defaultLanguage"),
            snippet.get("country"),
            getpath(payload, ["contentDetails", "relatedPlaylists", "uploads"]),
            statistics.get("viewCount"),
            statistics.get("hiddenSubscriberCount"),
            statistics.get("subscriberCount"),
            statistics.get("videoCount"),
            topic_ids,
            topic_categories,
            topic_keywords,
            status.get("privacyStatus"),
            status.get("madeForKids"),
            status.get("longUploadsStatus"),
            keywords,
            getpath(branding_settings, ["channel", "moderateComments"]),
            getpath(branding_settings, ["channel", "unsubscribedTrailer"]),
            getpath(branding_settings, ["image", "bannerExternalUrl"]),
        )



import sys
from typing import Union, List, Optional
from os import PathLike
from io import FileIO
from bs4 import BeautifulSoup
from urllib3 import Timeout
from ebbe import format_repr

# NOTE: yes this is repetitive, but mypy only understands this, don't ask questions...
if sys.version_info >= (3, 8):
    from typing import Literal, get_origin, get_type_hints, get_args
else:
    from typing_extensions import Literal, get_origin, get_type_hints, get_args

if sys.version_info >= (3, 10):
    from typing import TypeGuard, Concatenate, ParamSpec
else:
    from typing_extensions import TypeGuard, Concatenate, ParamSpec

if sys.version_info >= (3, 11):
    from typing import TypedDict, Required, NotRequired, Unpack
else:
    from typing_extensions import TypedDict, Required, NotRequired, Unpack

# Useful conditional imports
try:
    from urllib3 import HTTPHeaderDict
except ImportError:
    from urllib3._collections import HTTPHeaderDict

# Useful Any types
AnyPath = Union[str, PathLike]
AnyFileTarget = Union[AnyPath, FileIO]
AnyTimeout = Union[float, Timeout]

# Redirection types
RedirectionType = Literal[
    "hit",
    "location-header",
    "js-relocation",
    "refresh-header",
    "meta-refresh",
    "infer",
    "canonical",
]


class Redirection(object):
    __slots__ = ("status", "type", "url")

    status: Optional[int]
    url: str
    type: RedirectionType

    def __init__(
        self, url: str, _type: RedirectionType = "hit", status: Optional[int] = None
    ):
        self.status = status
        self.url = url
        self.type = _type

    def __repr__(self):
        return format_repr(self, ("type", "status", "url"))


RedirectionStack = List[Redirection]



# =============================================================================
# Minet Hyphe Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class HypheError(MinetError):
    pass


class HypheJSONRPCError(HypheError):
    def __init__(self, payload):
        super().__init__(payload["faultString"])
        self.code = payload["faultCode"]
        self.fault = payload["fault"]

    def __str__(self):
        return super().__str__() + ", Code: %i, Fault: %s" % (self.code, self.fault)


class HypheRequestFailError(HypheError):
    pass


class HypheCorpusAuthenticationError(HypheError):
    pass


class HypheCouldNotStartCorpusError(HypheError):
    pass



# =============================================================================
# Minet YouTube API Client
# =============================================================================
#
# A handy API client used by the CLI actions.
#
from typing import Deque, Tuple, Iterator, Any, Optional

import time
from ebbe import as_chunks
from collections import deque
from ural import urls_from_text, add_query_argument, is_url
from ebbe import getpath

from minet.web import (
    create_pool_manager,
    create_request_retryer,
    request,
    retrying_method,
)
from minet.loggers import sleepers_logger
from minet.youtube.utils import (
    ensure_video_id,
    ensure_channel_id,
    get_channel_main_playlist_id,
    seconds_to_midnight_pacific_time,
)
from minet.youtube.urls import YouTubeAPIURLFormatter
from minet.youtube.constants import (
    YOUTUBE_API_MAX_VIDEOS_PER_CALL,
    YOUTUBE_API_MAX_CHANNELS_PER_CALL,
    YOUTUBE_API_DEFAULT_SEARCH_ORDER,
    YOUTUBE_API_SEARCH_ORDERS,
)
from minet.youtube.exceptions import (
    YouTubeDisabledCommentsError,
    YouTubeNotFoundError,
    YouTubeInvalidAPIKeyError,
    YouTubeInvalidAPICallError,
    YouTubePotentiallyTransientInvalidAPICallError,
    YouTubeInvalidVideoTargetError,
    YouTubeInvalidChannelTargetError,
    YouTubeExclusiveMemberError,
    YouTubeUnknown403Error,
    YouTubeAccessNotConfiguredError,
)
from minet.youtube.types import (
    YouTubeVideo,
    YouTubeVideoSnippet,
    YouTubeComment,
    YouTubePlaylistVideoSnippet,
    YouTubeChannel,
)
from minet.youtube.scraper import YouTubeScraper


def get_channel_id(scraper: YouTubeScraper, channel_target: str) -> str:
    channel_id = ensure_channel_id(channel_target)

    if channel_id is not None:
        return channel_id

    if not is_url(channel_target):
        channel_target = "https://www.youtube.com/@" + channel_target.lstrip("@")

    channel_id = scraper.get_channel_id(channel_target)

    if channel_id is None:
        raise YouTubeInvalidChannelTargetError

    return channel_id


class YouTubeAPIClient(object):
    def __init__(self, key):
        if not isinstance(key, list):
            key = [key]

        self.keys = {k: True for k in key}
        self.current_key = key[0]
        self.pool_manager = create_pool_manager()
        self.scraper = YouTubeScraper()

        # YouTube's API is known to crash sometimes...
        self.retryer = create_request_retryer(
            retry_on_statuses=(503,),
            additional_exceptions=[YouTubePotentiallyTransientInvalidAPICallError],
        )
        self.url_formatter = YouTubeAPIURLFormatter()

    @retrying_method()
    def request_json(self, url):
        while True:
            final_url = add_query_argument(url, "key", self.current_key)
            response = request(final_url, pool_manager=self.pool_manager)
            data = response.json()

            if response.status == 403:
                if data is not None:
                    reason = getpath(data, ["error", "errors", 0, "reason"])

                    if reason == "accessNotConfigured":
                        msg = getpath(data, ["error", "message"])
                        url = next(urls_from_text(msg))
                        raise YouTubeAccessNotConfiguredError(msg, url=url)

                    elif reason == "commentsDisabled":
                        raise YouTubeDisabledCommentsError(url)

                    elif reason == "forbidden":
                        raise YouTubeExclusiveMemberError(url)

                    elif reason == "quotaExceeded":
                        # Current key is exhausted, disabling it and switching to another if there is one
                        if not self.rotate_key():
                            # If all keys are exhausted, start waiting until tomorrow and reset keys
                            sleep_time = seconds_to_midnight_pacific_time() + 10

                            sleepers_logger.warn(
                                "YouTube API limits reached for every key. Will now wait until midnight Pacific time!",
                                extra={
                                    "source": "YouTubeAPIClient",
                                    "sleep_time": 1,
                                },
                            )

                            time.sleep(sleep_time)

                            self.reset_keys()

                        continue

                raise YouTubeUnknown403Error

            if response.status == 404:
                raise YouTubeNotFoundError

            if response.status >= 400:
                error_message = ""

                if data is not None:
                    error_message = getpath(data, ["error", "message"], "").lower()

                if "api key not valid" in error_message:
                    raise YouTubeInvalidAPIKeyError

                if "transient" in error_message:
                    raise YouTubePotentiallyTransientInvalidAPICallError(
                        url, response.status, data
                    )

                raise YouTubeInvalidAPICallError(url, response.status, data)

            return data

    def rotate_key(self):
        self.keys[self.current_key] = False

        available_key = next(
            (key for key, available in self.keys.items() if available), None
        )

        if available_key:
            self.current_key = available_key
            return True

        self.current_key = None
        return False

    def reset_keys(self):
        for key in self.keys:
            if self.current_key is None:
                self.current_key = key

            self.keys[key] = True

    def channels(
        self, channels_target, key=None, raw=False
    ) -> Iterator[Tuple[Any, Optional[YouTubeChannel]]]:
        # TODO: we could chunk per not None
        for group in as_chunks(YOUTUBE_API_MAX_CHANNELS_PER_CALL, channels_target):
            group_data = []

            for item in group:
                target = key(item) if key is not None else item

                try:
                    channel_id = get_channel_id(self.scraper, target)
                    group_data.append((channel_id, item))
                except YouTubeInvalidChannelTargetError:
                    group_data.append((None, item))

            ids = [channel_id for channel_id, _ in group_data if channel_id is not None]

            url = self.url_formatter.channels(ids)

            result = self.request_json(url)

            indexed_result = {}

            # NOTE: as per #753, the "items" key is sometimes absent of the
            # API payload, e.g. if all the enquired channels have been
            # terminated for TOS violations.
            for item in result.get("items", []):
                channel_id = item["id"]

                if not raw:
                    item = YouTubeChannel.from_payload(item)

                indexed_result[channel_id] = item

            for channel_id, item in group_data:
                yield item, indexed_result.get(channel_id)

    def videos(
        self, videos, key=None, raw=False
    ) -> Iterator[Tuple[Any, Optional[YouTubeVideo]]]:
        # TODO: we could chunk per not None
        for group in as_chunks(YOUTUBE_API_MAX_VIDEOS_PER_CALL, videos):
            group_data = []

            for item in group:
                target = key(item) if key is not None else item
                video_id = ensure_video_id(target)
                group_data.append((video_id, item))

            ids = [video_id for video_id, _ in group_data if video_id is not None]

            url = self.url_formatter.videos(ids)

            result = self.request_json(url)

            indexed_result = {}

            for item in result["items"]:
                video_id = item["id"]

                if not raw:
                    item = YouTubeVideo.from_payload(item)

                indexed_result[video_id] = item

            for video_id, item in group_data:
                yield item, indexed_result.get(video_id)

    def search(
        self, query, order=YOUTUBE_API_DEFAULT_SEARCH_ORDER, raw=False
    ) -> Iterator[YouTubeVideoSnippet]:
        if order not in YOUTUBE_API_SEARCH_ORDERS:
            raise TypeError('unknown search order "%s"' % order)

        def generator():
            token = None

            while True:
                url = self.url_formatter.search(query, order=order, token=token)

                result = self.request_json(url)

                token = result.get("nextPageToken")

                for item in result["items"]:
                    if not raw:
                        item = YouTubeVideoSnippet.from_payload(item)

                    yield item

                if token is None or len(result["items"]) == 0:
                    break

        return generator()

    def comments(
        self, video_target, raw=False, full_replies=True
    ) -> Iterator[YouTubeComment]:
        video_id = ensure_video_id(video_target)

        if video_id is None:
            raise YouTubeInvalidVideoTargetError

        def generator():
            starting_url = self.url_formatter.comments(video_id)

            queue: Deque[Tuple[bool, str, str]] = deque(
                [(False, video_id, starting_url)]
            )

            while len(queue) != 0:
                is_reply, item_id, url = queue.popleft()

                result = self.request_json(url)

                for item in result["items"]:
                    comment_id = item["id"]
                    replies = getpath(item, ["replies", "comments"], [])
                    total_reply_count = getpath(item, ["snippet", "totalReplyCount"], 0)

                    if not raw:
                        item = (
                            YouTubeComment.from_parent_comment_payload(item)
                            if not is_reply
                            else YouTubeComment.from_reply_payload(
                                item, video_id=video_id
                            )
                        )

                    yield item

                    if is_reply:
                        continue

                    # Getting replies
                    if not full_replies or len(replies) >= total_reply_count:
                        for reply in replies:
                            if not raw:
                                reply = YouTubeComment.from_reply_payload(reply)

                            yield reply
                    elif total_reply_count > 0:
                        replies_url = self.url_formatter.replies(comment_id)

                        queue.append((True, comment_id, replies_url))

                # Next page
                token = result.get("nextPageToken")

                if token is not None and len(result["items"]) != 0:
                    forge = (
                        self.url_formatter.replies
                        if is_reply
                        else self.url_formatter.comments
                    )

                    next_url = forge(item_id, token=token)

                    queue.append((is_reply, item_id, next_url))

        return generator()

    def channel_videos(
        self,
        channel_target: str,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
    ) -> Iterator[YouTubePlaylistVideoSnippet]:
        channel_id = get_channel_id(self.scraper, channel_target)

        playlist_id = get_channel_main_playlist_id(channel_id)

        def generator():
            token = None

            while True:
                url = self.url_formatter.playlist_videos(playlist_id, token=token)

                result = self.request_json(url)

                token = result.get("nextPageToken")

                for item in result["items"]:
                    item = YouTubePlaylistVideoSnippet.from_payload(item)

                    if start_time and start_time > item.published_at:
                        break

                    elif end_time is None or item.published_at < end_time:
                        yield item

                if token is None or len(result["items"]) == 0:
                    break

        return generator()



# =============================================================================
# Minet Instagram Post-infos CLI Action
# =============================================================================
#
# Logic of the `instagram post-infos` action.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.instagram.utils import with_instagram_fatal_errors
from minet.instagram import InstagramAPIScraper
from minet.instagram.types import InstagramPost
from minet.instagram.exceptions import InstagramInvalidTargetError


@with_instagram_fatal_errors
@with_enricher_and_loading_bar(
    headers=InstagramPost, title="Scraping infos", unit="posts"
)
def action(cli_args, enricher, loading_bar):
    client = InstagramAPIScraper(cookie=cli_args.cookie)

    for i, row, user in enricher.enumerate_cells(
        cli_args.column, with_rows=True, start=1
    ):
        with loading_bar.step():
            try:
                result = client.post_infos(user)

                enricher.writerow(row, result)

            except InstagramInvalidTargetError:
                loading_bar.print(
                    "Given post (line %i) is probably not an Instagram post: %s"
                    % (i, user)
                )



from typing import Optional, List, Any

from dataclasses import dataclass
from casanova import TabularRecord, tabular_field

from minet.crowdtangle.constants import (
    CROWDTANGLE_STATISTICS,
    CROWDTANGLE_REACTION_TYPES,
    CROWDTANGLE_FULL_STATISTICS,
)


@dataclass
class CrowdTangleAccount(TabularRecord):
    ct_id: str
    id: str
    platform: str
    name: str
    handle: Optional[str]
    profile_image: str
    subscriber_count: int
    url: str
    verified: bool
    type: str
    page_admin_top_country: Optional[str]


def map_key(key, target):
    return [item[key] for item in target]


@dataclass
class CrowdTanglePost(TabularRecord):
    ct_id: str
    id: str
    platform: str
    type: str
    title: Optional[str]
    caption: Optional[str]
    message: Optional[str]
    description: Optional[str]
    date: str
    datetime: str
    updated: str
    link: str
    post_url: str
    score: float
    video_length_ms: str
    live_video_status: str
    actual_like_count: int
    expected_like_count: int
    actual_share_count: int
    expected_share_count: int
    actual_favorite_count: int
    expected_favorite_count: int
    actual_comment_count: int
    expected_comment_count: int
    actual_love_count: int
    expected_love_count: int
    actual_wow_count: int
    expected_wow_count: int
    actual_haha_count: int
    expected_haha_count: int
    actual_sad_count: int
    expected_sad_count: int
    actual_angry_count: int
    expected_angry_count: int
    actual_thankful_count: int
    expected_thankful_count: int
    actual_care_count: int
    expected_care_count: int
    account: CrowdTangleAccount
    links: List[str]
    expanded_links: List[str]
    media: Any = tabular_field(as_json=True)

    @classmethod
    def from_payload(cls, payload) -> "CrowdTanglePost":
        row = [
            payload["id"],
            payload["platformId"],
            payload["platform"],
            payload["type"],
            payload.get("title"),
            payload.get("caption"),
            payload.get("message"),
            payload.get("description"),
            payload["date"].split(" ", 1)[0],
            payload["date"],
            payload["updated"],
            payload.get("link"),
            payload.get("postUrl"),
            payload["score"],
            payload.get("videoLengthMS"),
            payload.get("liveVideoStatus"),
        ]

        stats = payload["statistics"]
        actual_stats = stats["actual"]
        expected_stats = stats["expected"]

        for name in CROWDTANGLE_STATISTICS:
            key = "%sCount" % name

            row.append(actual_stats.get(key, ""))
            row.append(expected_stats.get(key, ""))

        account = payload["account"]
        account = CrowdTangleAccount(
            account["id"],
            account.get("platformId"),
            account.get("platform"),
            account["name"],
            account.get("handle"),
            account.get("profileImage"),
            account["subscriberCount"],
            account["url"],
            account["verified"],
            account.get("accountType"),
            account.get("pageAdminTopCountry"),
        )

        row.extend(
            [
                # Account
                account,
                # Remaining
                map_key("original", payload.get("expandedLinks", [])),
                map_key("expanded", payload.get("expandedLinks", [])),
                payload.get("media"),
            ]
        )

        return cls(*row)


@dataclass
class CrowdTangleSummary(TabularRecord):
    angry_count: int
    comment_count: int
    haha_count: int
    like_count: int
    love_count: int
    sad_count: int
    share_count: int
    thankful_count: int
    wow_count: int

    @classmethod
    def from_payload(cls, payload) -> "CrowdTangleSummary":
        row = (payload["%sCount" % t] for t in CROWDTANGLE_REACTION_TYPES)
        return cls(*row)


@dataclass
class CrowdTangleLeaderboard(TabularRecord):
    ct_id: str
    name: str
    handle: str
    profile_image: str
    subscriber_count: int
    url: str
    verified: bool
    initial_subscriber_count: int
    final_subscriber_count: int
    subscriber_data_notes: str
    love_count: int
    wow_count: int
    thankful_count: int
    interaction_rate: int
    like_count: int
    haha_count: int
    comment_count: int
    share_count: int
    sad_count: int
    angry_count: int
    post_count: int
    total_interaction_count: int
    total_video_time_ms: int
    three_plus_minute_video_count: int

    @classmethod
    def from_payload(cls, payload) -> "CrowdTangleLeaderboard":
        account = payload["account"]
        subscriber_data = payload["subscriberData"]

        row = [
            account["id"],
            account["name"],
            account.get("handle"),
            account.get("profileImage"),
            account["subscriberCount"],
            account["url"],
            account["verified"],
            subscriber_data["initialCount"],
            subscriber_data["finalCount"],
            subscriber_data.get("notes"),
        ]

        summary = payload["summary"]

        for key, _ in CROWDTANGLE_FULL_STATISTICS:
            row.append(summary.get(key))

        return cls(*row)


@dataclass
class CrowdTangleList(TabularRecord):
    id: str
    title: str
    type: str

    @classmethod
    def from_payload(cls, payload) -> "CrowdTangleList":
        return cls(payload["id"], payload["title"], payload["type"])



# =============================================================================
# Minet Facebook Utils
# =============================================================================
#
# Miscellaneous helpers used by the minet.facebook namespace.
#
from http.cookies import SimpleCookie

from minet.cookies import coerce_cookie_for_url_from_browser
from minet.facebook.constants import FACEBOOK_URL


def fix_cookie(cookie_string):
    cookie = SimpleCookie()
    cookie.load(cookie_string)

    # NOTE: those cookie items can rat you out
    cookie.pop("m_pixel_ratio", None)
    cookie.pop("wd", None)
    cookie.pop("fr", None)

    # NOTE: cookie mangling of the `locale` does not work anymore
    # cookie["locale"] = "en_US"

    return "; ".join(key + "=" + morsel.coded_value for key, morsel in cookie.items())


def grab_facebook_cookie(source):
    cookie = coerce_cookie_for_url_from_browser(source, FACEBOOK_URL + "/")

    if not cookie or "c_user=" not in cookie.lower():
        return None

    return fix_cookie(cookie)



# =============================================================================
# Minet Twitter CLI Utils
# =============================================================================
#
# Miscellaneous generic functions used throughout the twitter actions.
#
import re
from twitter import TwitterHTTPError
from functools import wraps

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.exceptions import InvalidArgumentsError, FatalError
from minet.twitter import TwitterAPIClient

CHARACTERS = re.compile(r"[A-Za-z_]")
NUMBERS = re.compile(r"[0-9]+")
TWITTER_SCREEN_NAME = re.compile(r"[a-zA-Z0-9_]{1,15}")

ITEMS_PER_PAGE = 1000


def validate_query_boundaries(cli_args):
    if cli_args.start_time is not None and cli_args.end_time is not None:
        if cli_args.end_time < cli_args.start_time:
            raise InvalidArgumentsError("--end-time should be after --start-time!")

    if cli_args.since_id and cli_args.until_id:
        if cli_args.until_id < cli_args.since_id:
            raise InvalidArgumentsError("--until-id should be greater than --since-id!")


def with_twitter_client(api_version=None):
    def decorate(action):
        @wraps(action)
        def wrapper(cli_args, *args, **kwargs):
            nonlocal api_version

            v2_flag = getattr(cli_args, "v2", False)

            if api_version is None:
                api_version = "1.1" if not v2_flag else "2"

            client = TwitterAPIClient(
                cli_args.access_token,
                cli_args.access_token_secret,
                cli_args.api_key,
                cli_args.api_secret_key,
                api_version=api_version,
            )

            return action(cli_args, *args, **{"client": client}, **kwargs)

        return wrapper

    return decorate


def make_twitter_action(method_name, csv_headers):
    @with_enricher_and_loading_bar(
        headers=csv_headers,
        enricher_type="batch",
        title="Retrieving %s" % method_name,
        unit="users",
        nested=True,
        sub_unit=method_name,
    )
    @with_twitter_client()
    def action(cli_args, client, enricher, loading_bar):
        resuming_state = None

        if cli_args.resume:
            resuming_state = cli_args.output.pop_state()

        for row, user in enricher.cells(cli_args.column, with_rows=True):
            with loading_bar.step(user):
                all_ids = []
                next_cursor = -1
                result = None

                if cli_args.v2:
                    next_cursor = None

                if resuming_state is not None and resuming_state.last_cursor:
                    next_cursor = int(resuming_state.last_cursor)

                if cli_args.v2:
                    if is_not_user_id(user):
                        raise FatalError(
                            "The column given as argument doesn't contain user ids, you have probably given user screen names as argument instead. With --v2, you can only use user ids to retrieve followers."
                        )

                    client_kwargs = {"max_results": ITEMS_PER_PAGE}

                elif cli_args.ids:
                    if is_not_user_id(user):
                        raise FatalError(
                            "The column given as argument doesn't contain user ids, you have probably given user screen names as argument instead. \nTry removing --ids from the command."
                        )

                    client_kwargs = {"user_id": user}

                else:
                    if is_probably_not_user_screen_name(user):
                        raise FatalError(
                            "The column given as argument probably doesn't contain user screen names, you have probably given user ids as argument instead. \nTry adding --ids to the command."
                        )
                        # force flag to add

                    client_kwargs = {"screen_name": user}

                while True:
                    skip_in_output = None

                    if resuming_state:
                        skip_in_output = resuming_state.values_to_skip
                        resuming_state = None

                    if not cli_args.v2:
                        client_kwargs["cursor"] = next_cursor

                        try:
                            result = client.call([method_name, "ids"], **client_kwargs)
                        except TwitterHTTPError as e:
                            if e.e.code == 404:
                                # The user does not exist
                                loading_bar.inc_stat("not-found", style="error")
                                break

                            elif e.e.code == 401:
                                # The user is protected
                                loading_bar.inc_stat("protected", style="warning")
                                break

                            raise FatalError(
                                "Received an unusual %i error. Your key is probably dead, sorry :'("
                                % e.e.code
                            )

                        if result is not None:
                            all_ids = result.get("ids", [])
                            next_cursor = result.get("next_cursor", 0)

                            loading_bar.nested_advance(len(all_ids))

                            batch = []

                            for user_id in all_ids:
                                if skip_in_output and user_id in skip_in_output:
                                    continue

                                batch.append([user_id])

                        else:
                            break

                    else:
                        if method_name == "friends":
                            method_name_v2 = "following"
                        else:
                            method_name_v2 = method_name

                        try:
                            result = client.call(
                                route=["users", user, method_name_v2], **client_kwargs
                            )
                        except TwitterHTTPError as e:
                            if e.e.code == 404:
                                # The user does not exist
                                loading_bar.inc_stat("not-found", style="error")
                                break

                            elif e.e.code == 401:
                                # The user is protected
                                loading_bar.inc_stat("protected", style="warning")
                                break

                            raise FatalError(
                                "Received an unusual %i error. Your key is probably dead, sorry :'("
                                % e.e.code
                            )

                        if result is not None and "data" in result:
                            batch = []

                            for follower_metadata in result["data"]:
                                user_id = follower_metadata["id"]

                                if skip_in_output and user_id in skip_in_output:
                                    continue
                                batch.append([user_id])

                            loading_bar.nested_advance(len(result["data"]))

                            if "next_token" in result["meta"]:
                                next_cursor = result["meta"]["next_token"]
                                client_kwargs["pagination_token"] = next_cursor
                            else:
                                next_cursor = None

                        else:
                            break

                    enricher.writebatch(row, batch, next_cursor or None)

                    if next_cursor is None or next_cursor == 0:
                        break

    return action


def is_not_user_id(item):
    return bool(re.match(CHARACTERS, item))


def is_probably_not_user_screen_name(item):
    matches = TWITTER_SCREEN_NAME.fullmatch(item)
    if matches:
        return bool(NUMBERS.fullmatch(item))
    return True



from casanova import IndexedResumer

from minet.cli.argparse import command, FolderStrategyType
from minet.cli.constants import DEFAULT_CONTENT_FOLDER, DEFAULT_SCREENSHOT_FOLDER
from minet.cli.exceptions import InvalidArgumentsError

# TODO: lazyloading issue
from minet.fs import FolderStrategy
from minet.constants import (
    COOKIE_BROWSERS,
    DEFAULT_THROTTLE,
    DEFAULT_FETCH_MAX_REDIRECTS,
    DEFAULT_RESOLVE_MAX_REDIRECTS,
)

COMMON_ARGUMENTS = [
    {
        "flag": "--domain-parallelism",
        "help": "Max number of urls per domain to hit at the same time.",
        "type": int,
        "default": 1,
    },
    {
        "flags": ["-t", "--threads"],
        "help": "Number of threads to use. Will default to a conservative number, based on the number of available cores. Feel free to increase.",
        "type": int,
    },
    {
        "flag": "--throttle",
        "help": "Time to wait - in seconds - between 2 calls to the same domain.",
        "type": float,
        "default": DEFAULT_THROTTLE,
    },
    {
        "flag": "--url-template",
        "help": "A template for the urls to fetch. Handy e.g. if you need to build urls from ids etc.",
    },
    {
        "flag": "--timeout",
        "help": "Maximum time - in seconds - to spend for each request before triggering a timeout. Defaults to ~30s.",
        "type": float,
    },
]

FETCH_RESOLVE_COMMON_ARGUMENTS = [
    *COMMON_ARGUMENTS,
    {
        "flags": ["-g", "--grab-cookies"],
        "help": 'Whether to attempt to grab cookies from your computer\'s browser (supports "firefox", "chrome", "chromium", "opera" and "edge").',
        "choices": COOKIE_BROWSERS,
    },
    {
        "flags": ["-H", "--header"],
        "help": "Custom headers used with every requests.",
        "action": "append",
        "dest": "headers",
    },
    {
        "flags": ["-k", "--insecure"],
        "help": "Whether to allow ssl errors when performing requests or not.",
        "action": "store_true",
    },
    {
        "flags": ["-X", "--request"],
        "help": "The http method to use. Will default to GET.",
        "dest": "method",
        "default": "GET",
    },
    {"flags": ["-x", "--proxy"], "help": "Proxy server to use."},
    {
        "flag": "--spoof-user-agent",
        "help": 'Whether to use a plausible random "User-Agent" header when making requests.',
        "action": "store_true",
    },
    {
        "flag": "--retries",
        "help": "Number of times to retry on timeout & common network-related issues.",
        "type": int,
        "default": 0,
    },
]

COMMON_IO_ARGUMENTS = [
    {
        "flags": ["-f", "--filename-column"],
        "help": 'Name of the column used to build retrieved file names. Defaults to a md5 hash of final url. If the provided file names have no extension (e.g. ".jpg", ".pdf", etc.) the correct extension will be added depending on the file type.',
    },
    {
        "flag": "--filename-template",
        "help": "A template for the name of the fetched files.",
    },
    {
        "flag": "--folder-strategy",
        "help": "Name of the strategy to be used to dispatch the retrieved files into folders to alleviate issues on some filesystems when a folder contains too much files. Note that this will be applied on top of --filename-template. All of the strategies are described at the end of this help.",
        "default": "flat",
        "type": FolderStrategyType(),
    },
]


def resolve_fetch_arguments(cli_args):
    if cli_args.dont_save:
        cli_args.contents_in_report = False

    if cli_args.contents_in_report and cli_args.compress_on_disk:
        raise InvalidArgumentsError(
            "Cannot both --compress-on-disk and get --contents-in-report!"
        )

    # --sqlar disables --compress-on-disk
    if cli_args.sqlar:
        cli_args.compress_on_disk = False


FETCH_COMMAND = command(
    "fetch",
    "minet.cli.fetch.fetch",
    title="Minet Fetch Command",
    description="""
        Use multiple threads to fetch batches of urls from a CSV file. The
        command outputs a CSV report with additional metadata about the
        HTTP calls and will generally write the retrieved files in a folder
        given by the user.
    """,
    epilog=f"""
        Columns being added to the output:

        . "original_index": index of the line in the original file (the output will be
            arbitrarily ordered since multiple requests are performed concurrently).
        . "resolved_url": final resolved url (after templating & solving redirects).
        . "http_status": HTTP status code of the request, e.g. 200, 404, 503 etc.
        . "datetime_utc": datetime when the response was finished.
        . "fetch_error": an error code if anything went wrong when performing the request.
        . "path": path to the downloaded file, relative to the folder given
            through -O/--output-dir.
        . "mimetype": detected mimetype of the requested file.
        . "encoding": detected encoding of the requested file if relevant.
        . "body_size": size of the downloaded document in bytes.
        . "body": if -c/--contents-in-report is set, will contain the
            downloaded text and the files won't be written to disk.

        --folder-strategy options:

        {FolderStrategy.DOCUMENTATION}

        Examples:

        . Fetching a single url (can be useful when piping):
            $ minet fetch "https://www.lemonde.fr"

        . Fetching a batch of url from existing CSV file:
            $ minet fetch url -i file.csv > report.csv

        . Piping to minet extract:
            $ minet fetch url -i file.csv -c | minet extract -i -

        . CSV input from stdin (mind the `-`):
            $ xsv select url file.csv | minet fetch url -i - > report.csv

        . Downloading files in specific output directory:
            $ minet fetch url -i file.csv -O html > report.csv
    """,
    resolve=resolve_fetch_arguments,
    resumer=IndexedResumer,
    variadic_input={"dummy_column": "url"},
    arguments=[
        *FETCH_RESOLVE_COMMON_ARGUMENTS,
        *COMMON_IO_ARGUMENTS,
        {
            "flags": ["-O", "--output-dir"],
            "help": "Directory where the fetched files will be written.",
            "default": DEFAULT_CONTENT_FOLDER,
        },
        {
            "flag": "--max-redirects",
            "help": "Maximum number of redirections to follow before breaking.",
            "type": int,
            "default": DEFAULT_FETCH_MAX_REDIRECTS,
        },
        {
            "flags": ["-z", "--compress-on-disk"],
            "help": "Whether to compress the contents.",
            "action": "store_true",
        },
        {
            "flag": "--compress-transfer",
            "help": 'Whether to send a "Accept-Encoding" header asking for a compressed response. Usually better for bandwidth but at the cost of more CPU work.',
            "action": "store_true",
        },
        {
            "flags": ["-c", "--contents-in-report"],
            "help": "Whether to include retrieved contents, e.g. html, directly in the report and avoid writing them in a separate folder. This requires to standardize encoding and won't work on binary formats.",
            "dest": "contents_in_report",
            "action": "store_true",
        },
        {
            "flags": ["-D", "--dont-save"],
            "help": "Use not to write any downloaded file on disk.",
            "action": "store_true",
        },
        {
            "flag": "--keep-failed-contents",
            "help": "Whether to keep & write contents for failed (i.e. non-200) http requests.",
            "action": "store_true",
        },
        {
            "flag": "--standardize-encoding",
            "help": "Whether to systematically convert retrieved text to UTF-8.",
            "action": "store_true",
        },
        {
            "flag": "--only-html",
            "help": (
                "Only download urls if they are not unambiguously not html (e.g. if the url has the .pdf or .jpg extension etc.). "
                "Then, if the url was downloaded, it will only be written on disk if the http body actually looks like html."
            ),
            "action": "store_true",
        },
        {
            "flag": "--pycurl",
            "help": "Whether to use the pycurl library to perform the calls.",
            "action": "store_true",
        },
        {
            "flag": "--sqlar",
            "help": "Whether to write files into a single-file sqlite archive rather than as individual files on the disk.",
            "action": "store_true",
        },
    ],
)

RESOLVE_COMMAND = command(
    "resolve",
    "minet.cli.fetch.fetch",
    title="Minet Resolve Command",
    description="""
        Use multiple threads to resolve batches of urls from a CSV file. The
        command outputs a CSV report with additional metadata about the
        HTTP calls and the followed redirections.
    """,
    epilog="""
        Columns being added to the output:

        . "original_index": index of the line in the original file (the output will be
            arbitrarily ordered since multiple requests are performed concurrently).
        . "resolved_url": final resolved url (after solving redirects).
        . "http_status": HTTP status code of the request, e.g. 200, 404, 503 etc.
        . "resolution_error": an error code if anything went wrong when performing the request.
        . "redirect_count": total number of redirections to reach the final url.
        . "redirect_chain": list of redirection types separated by "|".

        Examples:

        . Resolving a batch of url from existing CSV file:
            $ minet resolve url_column -i file.csv > report.csv

        . CSV input from stdin (mind the `-`):
            $ xsv select url_column file.csv | minet resolve url_column - > report.csv

        . Resolving a single url:
            $ minet resolve https://lemonde.fr
    """,
    resumer=IndexedResumer,
    variadic_input={"dummy_column": "url"},
    arguments=[
        *FETCH_RESOLVE_COMMON_ARGUMENTS,
        {
            "flag": "--max-redirects",
            "help": "Maximum number of redirections to follow before breaking.",
            "type": int,
            "default": DEFAULT_RESOLVE_MAX_REDIRECTS,
        },
        {
            "flag": "--follow-meta-refresh",
            "help": "Whether to follow meta refresh tags. Requires to buffer part of the response body, so it will slow things down.",
            "action": "store_true",
        },
        {
            "flag": "--follow-js-relocation",
            "help": "Whether to follow typical JavaScript window relocation. Requires to buffer part of the response body, so it will slow things down.",
            "action": "store_true",
        },
        {
            "flag": "--infer-redirection",
            "help": "Whether to try to heuristically infer redirections from the urls themselves, without requiring a HTTP call.",
            "action": "store_true",
        },
        {
            "flag": "--canonicalize",
            "help": "Whether to extract the canonical url from the html source code of the web page if found. Requires to buffer part of the response body, so it will slow things down.",
            "action": "store_true",
        },
        {
            "flag": "--only-shortened",
            "help": "Whether to only attempt to resolve urls that are probably shortened.",
            "action": "store_true",
        },
    ],
)

SCREENSHOT_COMMAND = command(
    "screenshot",
    "minet.cli.fetch.fetch",
    title="Minet Screenshot Command",
    description="""
        Use multithreaded browser emulation to screenshot batches of urls
        from a CSV file. The command outputs a CSV report with additional
        metadata about the HTTP calls and the produced screenshots on disk.
    """,
    epilog="""
        Columns being added to the output:

        . "original_index": index of the line in the original file (the output will be
            arbitrarily ordered since multiple requests are performed concurrently).
        . "http_status": HTTP status code of the request, e.g. 200, 404, 503 etc.
        . "screenshot_error": an error code if anything went wrong when performing the request.
        . "path": path to the downloaded file, relative to the folder given
            through -O/--output-dir.

        --folder-strategy options:

        {FolderStrategy.DOCUMENTATION}

        Examples:

        . Screenshot a batch of url from existing CSV file:
            $ minet screenshot url_column -i file.csv > report.csv

        . CSV input from stdin (mind the `-`):
            $ xsv select url_column file.csv | minet screenshot url_column - > report.csv

        . Screenshot a single url:
            $ minet screenshot https://lemonde.fr
    """,
    resumer=IndexedResumer,
    variadic_input={"dummy_column": "url"},
    arguments=[
        *COMMON_ARGUMENTS,
        *COMMON_IO_ARGUMENTS,
        {
            "flags": ["-O", "--output-dir"],
            "help": "Directory where the screenshots will be written.",
            "default": DEFAULT_SCREENSHOT_FOLDER,
        },
        {
            "flag": "--full-page",
            "help": "Whether to create full page screenshots.",
            "action": "store_true",
        },
        {
            "flag": "--width",
            "help": "Page width in pixels.",
            "type": int,
            "default": 1920,
        },
        {
            "flag": "--height",
            "help": "Page height in pixels.",
            "type": int,
            "default": 1080,
        },
        {
            "flag": "--wait-until",
            "help": "What to wait to consider the web page loaded.",
            "default": "load",
            "choices": ["domcontentloaded", "load", "networkidle", "commit"],
        },
        {
            "flag": "--adblock",
            "help": "Whether to use the ublock-origin browser extension.",
            "action": "store_true",
        },
        {
            "flag": "--automatic-consent",
            "help": 'Whether to use the "I still don\'t care about cookies" browser extension to try and get rid of GDPR/cookies consent forms when screenshotting.',
            "action": "store_true",
        },
        {
            "flag": "--wait",
            "help": "Time to wait, in seconds, before taking the screenshot. This might be a good idea if none of the --wait-until strategies work for you and if you need to give more time to the browser extensions to kick in. This obviously makes the whole process slower.",
            "type": float,
        },
    ],
)



# =============================================================================
# Minet Telegram CLI Action
# =============================================================================
#
# Logic of the `tl` action.
#
from minet.cli.argparse import command

# TODO: this is a lazyloading problem
from minet.telegram.constants import TELEGRAM_DEFAULT_THROTTLE

THROTTLE_ARGUMENT = {
    "flag": "--throttle",
    "help": "Throttling time, in seconds, to wait between each request.",
    "type": float,
    "default": TELEGRAM_DEFAULT_THROTTLE,
}

TELEGRAM_CHANNEL_INFOS_SUBCOMMAND = command(
    "channel-infos",
    "minet.cli.telegram.channel_infos",
    title="Minet Telegram Channel-Infos Command",
    description="""
        Scrape a Telegram channel's infos.
    """,
    epilog="""
        Examples:
        . Scraping a channel's infos:
            $ minet telegram channel-infos nytimes > infos.csv
    """,
    variadic_input={
        "dummy_column": "channel_name",
        "item_label": "channel name / url",
        "item_label_plural": "channel names / urls",
    },
    arguments=[THROTTLE_ARGUMENT],
)

TELEGRAM_CHANNEL_MESSAGES_SUBCOMMAND = command(
    "channel-messages",
    "minet.cli.telegram.channel_messages",
    title="Minet Telegram Channel-Messages Command",
    description="""
        Scrape Telegram channel messages.
    """,
    epilog="""
        Examples:
        . Scraping a group's posts:
            $ minet telegram channel-messages nytimes > messages.csv
    """,
    variadic_input={
        "dummy_column": "channel_name",
        "item_label": "channel name / url",
        "item_label_plural": "channel names / urls",
    },
    arguments=[THROTTLE_ARGUMENT],
)

TELEGRAM_COMMAND = command(
    "telegram",
    "minet.cli.mediacloud",
    aliases=["tl"],
    title="Minet Telegram Command",
    description="""
        Collects data from Telegram.
    """,
    subcommands=[
        TELEGRAM_CHANNEL_INFOS_SUBCOMMAND,
        TELEGRAM_CHANNEL_MESSAGES_SUBCOMMAND,
    ],
)



import casanova
from os.path import isdir

from minet.crawl.queue import CrawlerQueue, CrawlerQueueRecord

from minet.cli.exceptions import FatalError


def action(cli_args):
    if not isdir(cli_args.queue_dir):
        raise FatalError("queue does not exist")

    queue = CrawlerQueue(cli_args.queue_dir, inspect=True)

    writer = casanova.writer(
        cli_args.output, fieldnames=CrawlerQueueRecord.fieldnames()
    )

    for record in queue.dump():
        writer.writerow(record)
